<?php

declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: ba441a140530 $
 * $Author: Marc Kraemer <marc.kraemer@webedition.org> $
 * $Date: 2021-01-25 18:28 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */
class we_backup_exportZIP extends we_backup_exportBase implements Serializable{

	const INFO_FILE = 'info';
	const MAP_FILE = 'tableMap';
	const EXPORT_TYPE = 'backup_ZIP_JSON';

	/** @var ?ZipArchive */
	private $fh = null;

	/** @var int */
	private $currentFile = 0;

	public function __construct(string $filename){
		parent::__construct($filename . '.zip');
		we_base_file::delete($this->backup_file);
		$this->open(true);
		$this->fh->addEmptyDir(self::FILES_DIR);
		//create base structure
		/* $this->fh->addEmptyDir(self::SITE_DIR);
		  $this->fh->addEmptyDir(self::EXPORT_DIR);
		  $this->fh->addEmptyDir(self::TEMPLATES_DIR);
		  $this->fh->addEmptyDir(self::CONF_DIR);
		  $this->fh->addEmptyDir(self::HOOK_DIR);
		  $this->fh->addEmptyDir(self::TAG_DIR);
		  $this->fh->addEmptyDir(self::EXTERNALS_DIR); */
	}

	public function open(bool $init = false): bool{
		if($this->fh){
			return true;
		}
		if($init && file_exists($this->backup_file)){//make sure we have a new file
			we_base_file::delete($this->backup_file);
		}
		$this->fh = new ZipArchive();
		$ret = $this->fh->open($this->backup_file, ($init ? ZipArchive::CREATE : 0));
		if($ret !== true){
			$this->fh = null;
		}
		return $ret === true;
	}

	public function writeBackupDataToFile(string $data): bool{
		$this->open();
		$this->fh->addFromString(self::EXPORT_DIR . '/' . sprintf('%06d', ++$this->currentFile) . '.json', $data);
		$arr = array_keys($this->tableMap);
		$table = end($arr);

		we_backup_util::addLog('save to ' . $this->currentFile . ' :' . $table . ' stat (' . $this->currentFile . ',' . $this->binFiles . ')');
		return true;
	}

	public function writeBinFile(string $file, string $dir = self::FILES_DIR, string $as = ''): bool{
		$this->open();
		$filename = WE_DOCUMENT_ROOT . $file;
		if(is_file($filename) && is_readable($filename)){
			$this->binFiles++;
			$this->fh->addFile($filename, $dir . we_base_file::addPreSlash($as ?: $file));
			return true;
		}
		return false;
	}

	public function writeBinWEFile(int $id, string $type = we_base_ContentTypes::BACKUP_BINARY): bool{
		$bin = we_exim_contentProvider::getInstance($type, $id);
		'@phan-var we_backup_binary|we_backup_version $bin';
		if($bin instanceof we_backup_version){
			$path = WE_VERSION_DIR . $bin->Path;
			if($bin->Path && file_exists(WE_DOCUMENT_ROOT . $path)){
				$this->writeBinFile($path, self::VERSIONS_DIR, $bin->Path);
				return $this->writeBackupData($bin->getBackup());
			}
			return false;
		}

		$path = $bin->Path;
		if(($docExists = file_exists(($docPath = WE_DOCUMENT_ROOT . $path)))){
			$this->writeBinFile($path, self::FILES_DIR, $bin->Path);
		}
		if(file_exists(($sitePath = WE_DOCUMENT_ROOT . WE_SITE_DIR . $bin->Path)) && (!$docExists || //if only site exists
				filesize($docPath) != filesize($sitePath) ||
				sha1_file($docPath) != sha1_file($sitePath)
				)
		){
			$this->writeBinFile(WE_SITE_DIR . $bin->Path, self::SITE_DIR, $bin->Path);
		}
		return $this->writeBackupData($bin->getBackup());
	}

	public function close(): bool{
		if(!$this->fh){
			return false;
		}
		parent::close();
		$ret = $this->fh->close();
		$this->fh = null;
		return $ret;
	}

	public function finish(): void{
		if(!$this->fh){
			$this->open();
		}
		$settings = we_backup_settings::i();
		//@phan-suppress-next-line PhanPossiblyFalseTypeArgumentInternal
		$fileInfo = new we_backup_fileInfo();
		$fileInfo->setData($settings->comment, $this->currentFile, $this->binFiles);
		$this->fh->addFromString(self::INFO_FILE, json_encode($fileInfo, JSON_UNESCAPED_UNICODE | JSON_PRETTY_PRINT) ?: '');
		$this->fh->addFromString(self::MAP_FILE, json_encode($this->tableMap, JSON_UNESCAPED_UNICODE | JSON_PRETTY_PRINT) ?: '');
		$this->close();
		parent::finish();
	}

	public function flush(bool $force = false): void{
		//we don't flush
		//this will issue the next backup file to be opened
		if($force || strlen($this->buffer) > self::MAX_BUFFER){//if a new table is issued
			parent::close();
		}
	}

	protected function getOffset(): int{
		//we have not written buffer, so the index is not correct
		return $this->currentFile + 1;
	}

	public function serialize(): string{
		$this->close();
		return serialize([$this->currentFile, $this->backup_file, $this->binFiles, $this->tableMap]);
	}

	public function unserialize($serialized): void{
		[$this->currentFile, $this->backup_file, $this->binFiles, $this->tableMap] = unserialize($serialized);
	}

}
