<?php

declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: 12321095aae8 $
 * $Author: Marc Kraemer <marc.kraemer@webedition.org> $
 * $Date: 2021-01-26 01:01 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */
class we_backup_importZIP extends we_backup_importBase implements Serializable{

	/** @var ?ZipArchive */
	private $fh = null;

	/** @var int */
	private $currentFile = 0;

	/** @var int */
	private $binFiles = 0;

	/** @var we_backup_fileInfo */
	private $info;

	/** @var array<string,int> */
	private $skippedFiles = [];

	/** @var array<string,array> */
	private $otherFiles = [
		we_backup_exportBase::CONF_DIR => [],
		we_backup_exportBase::EXTERNALS_DIR => [],
		we_backup_exportBase::HOOK_DIR => [],
		we_backup_exportBase::TAG_DIR => []
	];

	/** @var array<string,array> */
	private $tableMap = [];

	public function __construct(string $filename = ''){
		parent::__construct($filename);
	}

	public function open(): bool{
		if($this->fh){
			return true;
		}
		$this->fh = new ZipArchive();
		return $this->fh->open($this->backup_file) === true;
	}

	public function close(): bool{
		if($this->fh){
			$this->fh->close();
			$this->fh = null;
			return true;
		}
		return false;
	}

	private function realRead(): string{
		$table = '';
		$next = [];
		foreach($this->tableMap as $tab => $data){
			if($table){
				$next = $data;
				break;
			}
			if(key($data) == $this->currentFile){
				$table = $tab;
			}
		}

		if($table && we_backup_util::getRealTableName($table) === ''){//skip table
			$this->currentFile = (int) key($next) - 1; //note this is incremented on next request
			$this->binFiles = current($next);
			we_backup_util::addLog('skip table ' . $table . ' next file ' . ($this->currentFile + 1));
			we_backup_util::writeLog();

			return '';
		}

		$data = $this->fh->getFromName(we_backup_exportZIP::EXPORT_DIR . '/' . sprintf('%06d', $this->currentFile) . '.json');
		$settings = we_backup_settings::i();
		$settings->offset = $this->currentFile + $this->binFiles;
		we_backup_util::addLog('read (' . $settings->offset . '/' . $settings->offset_end . ') from ' . $this->currentFile . ($table ? ': ' . $table : ''));
		return $data ?: '';
	}

	public function readLines(): string{
		if(!$this->open()){
			throw new we_exception_file('unable to open file');
		}
		if(++$this->currentFile <= $this->info->BackupFiles){
			return $this->realRead();
		}
		$settings = we_backup_settings::i();
//handle tag
		if($this->otherFiles[we_backup_exportBase::TAG_DIR]){
			we_backup_util::addLog('handle tags');
			foreach($this->otherFiles[we_backup_exportBase::TAG_DIR] as $curFile){
				[$dir, $file] = explode('/', $curFile, 2);
				$ret = $this->realExtract(we_backup_exportBase::TAG_DIR . '/', $curFile, WE_CUSTOM_PATH . ($dir === 'code' ? we_weTag_tag::DIR : we_weTag_twbase::DIR));
				we_backup_util::addLog('tag ' . $dir . ' ' . $file);
				if(!$ret){
					we_backup_util::addLog('FAILED to extract tag ' . we_backup_exportBase::TAG_DIR . '/'. $file);
				}
			}
			$this->otherFiles[we_backup_exportBase::TAG_DIR] = [];
			$settings->offset = $this->currentFile + $this->binFiles;
			we_backup_util::addLog('handled tags stat(' . $settings->offset . '/' . $settings->offset_end . ')', true);
		}
//handle hook
		if($this->otherFiles[we_backup_exportBase::HOOK_DIR]){
			we_backup_util::addLog('handle hooks');
			foreach($this->otherFiles[we_backup_exportBase::HOOK_DIR] as $file){
				$this->realExtract(we_backup_exportBase::HOOK_DIR . '/', $file, WE_CUSTOM_PATH . we_hook_handler::DIR);
				we_backup_util::addLog('hook ' . $file);
			}
			$this->otherFiles[we_backup_exportBase::HOOK_DIR] = [];
			$settings->offset = $this->currentFile + $this->binFiles;
			we_backup_util::addLog('handled hooks stat(' . $settings->offset . '/' . $settings->offset_end . ')', true);
		}
		//handle extern
		if($this->otherFiles[we_backup_exportBase::EXTERNALS_DIR]){
			we_backup_util::addLog('handle external files (left: ' . count($this->otherFiles[we_backup_exportBase::EXTERNALS_DIR]) . ')');
			$cnt = 0;
			foreach($this->otherFiles[we_backup_exportBase::EXTERNALS_DIR] as $key => $file){
				if($cnt++ > 50){//
					$settings->offset = $this->currentFile + $this->binFiles;
					return '';
				}
				$this->realExtract(we_backup_exportBase::EXTERNALS_DIR . '/', $file, (string) WE_DOCUMENT_ROOT);
				we_backup_util::addLog('extern ' . $file);
				unset($this->otherFiles[we_backup_exportBase::EXTERNALS_DIR][$key]);
			}
			$settings->offset = $this->currentFile + $this->binFiles;
			return '';
		}

		//handle config last, it may cause session reset
		if($this->otherFiles[we_backup_exportBase::CONF_DIR]){
			we_backup_util::addLog('handle config');
			$files = we_backup_util::getSettingsFiles(false);
			foreach($this->otherFiles[we_backup_exportBase::CONF_DIR] as $file){
				'@phan-var string $file';
				if(array_key_exists($file, $files)){
					we_backup_util::addLog('config handle ' . $file);
					switch($file){
						case 'we_conf_global.inc.php':
							if($this->realExtract(we_backup_exportBase::CONF_DIR . '/', $file, TEMP_PATH)){ //file found in backup
								we_base_preferences::check_global_config(true, TEMP_PATH . $file, [
									'SYSTEM_WE_SESSION_CRYPT'//keep this setting since session would be killed otherwise
								]);
								we_base_file::delete(TEMP_PATH . $file);
							}
							break;
						default:
							$this->realExtract(we_backup_exportBase::CONF_DIR . '/', $file, dirname(WE_DOCUMENT_ROOT . $files[$file]));
					}
				} else {
					we_backup_util::addLog('config skip (not used) ' . $file);
					//increase but don't install
					$this->binFiles++;
				}
			}
			$this->otherFiles[we_backup_exportBase::CONF_DIR] = [];
			$settings->offset = $this->currentFile + $this->binFiles;
			we_backup_util::addLog('handled config stat(' . $settings->offset . '/' . $settings->offset_end . ')', true);

			return '';
		}
		$settings->offset++;
		return '';
	}

	private function realExtract(string $basedir, string $filename, string $destDir): bool{
		if(!$this->open()){
			throw new we_exception_file('unable to open file');
		}

		if(!$this->fh->locateName($basedir . $filename)){
			//do not increase, since if file not found, it was not counted before
			return false;
		}
		$path = dirname($filename);
		$path = in_array($path, ['.', '/']) ? '' : $path;
		$settings = we_backup_settings::i();

		if($path !== $filename && !file_exists($destDir . $path)){
			we_backup_util::addLog('create dir ' . $destDir . $path);
			we_base_file::createLocalFolderByPath($destDir . $path);
		}
		//extractTo does not work since there is no option to extract /files/a.txt to path/a.txt - it will always want to use path/files/a.txt
		if(!copy('zip://' . $this->backup_file . '#' . $basedir . $filename, $destDir . $filename)){
			we_backup_util::addLog('failed to extract ' . $basedir . $filename . ' to ' . $destDir . $path);
		} else {
			we_backup_util::addLog('extract (' . $settings->offset . '/' . $settings->offset_end . '): ' . $filename/* . ' to ' . $destDir . $path */);
		}
		$this->binFiles++;
		$settings->offset = $this->currentFile + $this->binFiles;
		return true;
	}

	public function extractBinFile(string $dir, string $file): void{
		$settings = we_backup_settings::i();
		switch($dir){
			case we_backup_exportBase::FILES_DIR:
				if(str_contains($file, WEBEDITION_DIR)){
					//no write to we dir!!
					return;
				}
				if($settings->option(we_backup_settings::OPTION_BACKUP_BINARY)){
					$this->realExtract(we_backup_exportBase::FILES_DIR, $file, (string) WE_DOCUMENT_ROOT);
					$this->realExtract(we_backup_exportBase::SITE_DIR, $file, WE_DOCUMENT_ROOT . WE_SITE_DIR);
				} else {
					($this->fh->locateName(we_backup_exportBase::FILES_DIR . $file)) && $this->binFiles++;
					($this->fh->locateName(we_backup_exportBase::SITE_DIR . $file)) && $this->binFiles++;
				}
				break;
			case we_backup_exportBase::VERSIONS_DIR:
				if(!$file){
					break;
				}
				//we_backup_util::addLog('version for ' . $file);

				if($settings->isHandled(we_backup_settings::OPTION_BACKUP_VERSION_BIN)){
					$this->realExtract(we_backup_exportBase::VERSIONS_DIR . '/', $file, WE_VERSION_PATH);
				} else {
					$this->binFiles++;
				}
				break;
			default:
				t_e('unknown source dir', $dir, $file);
		}
		$settings->offset = $this->currentFile + $this->binFiles;

		return;
	}

	public function canReadMore(): bool{
		$settings = we_backup_settings::i();
		return $settings->offset < $settings->offset_end;
	}

	public function getOnlyInfo(): we_backup_fileInfo{
		if(!$this->open()){
			throw new we_exception_file('unable to open file');
		}

		$info = new we_backup_fileInfo();
		$info->setFromJson($this->fh->getFromName(we_backup_exportZIP::INFO_FILE) ?: '');

		$this->close();
		return $info;
	}

	public function getInfo(): we_backup_fileInfo{
		if(!$this->open()){
			throw new we_exception_file('unable to open file');
		}

		$settings = we_backup_settings::i();
//setup all external files, conffiles
		$this->skippedFiles = [
			we_backup_exportBase::CONF_DIR => 0,
			we_backup_exportBase::EXTERNALS_DIR => 0,
			we_backup_exportBase::HOOK_DIR => 0,
			we_backup_exportBase::TAG_DIR => 0,
		];
		for($i = 0; $i < $this->fh->numFiles; $i++){
			$hash = explode('/', $this->fh->getNameIndex($i) ?: '/', 2);
			if(count($hash) < 2){//files in root
				continue;
			}
			[$dir, $file] = $hash;
			switch($dir){
				case we_backup_exportBase::CONF_DIR:
					if($settings->isHandled(we_backup_settings::OPTION_BACKUP_SETTINGS)){
						$this->otherFiles[$dir][] = $file;
					} else {
						$this->skippedFiles[$dir]++;
					}
					break;
				case we_backup_exportBase::EXTERNALS_DIR:
					if($settings->option(we_backup_settings::OPTION_BACKUP_EXTERN)){
						$this->otherFiles[$dir][] = $file;
					} else {
						$this->skippedFiles[$dir]++;
					}
					break;
				case we_backup_exportBase::HOOK_DIR:
					if($settings->isHandled(we_backup_settings::OPTION_BACKUP_HOOKS)){
						$this->otherFiles[$dir][] = $file;
					} else {
						$this->skippedFiles[$dir]++;
					}
					break;
				case we_backup_exportBase::TAG_DIR:
					if($settings->isHandled(we_backup_settings::OPTION_BACKUP_CUSTOMTAGS)){
						$this->otherFiles[$dir][] = $file;
					} else {
						$this->skippedFiles[$dir]++;
					}
					break;
			}
		}
		we_backup_util::addLog('Skipped files: ' . print_r($this->skippedFiles, true));
		//add skipped files
		$this->binFiles += array_sum($this->skippedFiles);

		$this->tableMap = we_unserialize($this->fh->getFromName(we_backup_exportZIP::MAP_FILE) ?: '');
		$this->info = new we_backup_fileInfo();
		$this->info->setFromJson($this->fh->getFromName(we_backup_exportZIP::INFO_FILE) ?: '');
		return $this->info;
	}

	public function serialize(): string{
		$this->close();
		return serialize([$this->currentFile, $this->backup_file, $this->binFiles, $this->otherFiles, $this->info, $this->tableMap, $this->skippedFiles]);
	}

	public function unserialize($serialized): void{
		[$this->currentFile, $this->backup_file, $this->binFiles, $this->otherFiles, $this->info, $this->tableMap, $this->skippedFiles] = unserialize($serialized);
	}

}
