<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: 19e3d3aaf587 $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2020-12-15 23:51 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */
abstract class we_backup_util{
	public const backupMarker = '<!-- webackup -->';
	public const weXmlExImHead = '<webEdition';
	public const weXmlExImFooter = '</webEdition>';
	public const weXmlExImProtectCode = '<?php exit();?>'; //FIXME:remove
	public const SAVE_DIR = ''; //skipped
	public const LOG_FILE = self::SAVE_DIR . 'lastlog.php';
	public const FILE_PREFIX = 'weBackup_';

	static $logCnt = 0;

	/**
	 *
	 * @param string $table
	 * @param int $execTime
	 * @return boolean true, if still time left
	 */
	public static function limitsReached(string $table, int $execTime, int $memMulti = 2): bool{
		$settings = we_backup_settings::i();
		'@phan-var we_backup_settings $settings';
		if($table){
//check if at least 10 avg rows
			$rowSz = $settings->limits['avgLen'][strtolower(stripTblPrefix($table))];
			if(memory_get_usage(true) + 10 * $rowSz > $settings->limits['mem']){
				return false;
			}
		} elseif($settings->limits['lastMem'] != 0){
			$cur = memory_get_usage(true);
			$diff = $cur - $settings->limits['lastMem'];
			if($cur + $diff * $memMulti > $settings->limits['mem']){
				return false;
			}
		}
		$settings->limits['lastMem'] = memory_get_usage(true);

		if($execTime == 0){
			$execTime = 0.2;
		}

		$maxTime = $settings->limits['exec'] - 2;
		if(time() - ((int) $settings->limits['requestTime']) + 2 * $execTime > $maxTime){
			return false;
		}

		return true;
	}

	/**
	 *
	 * @param bool $import
	 * @return array<string,string>
	 */
	public static function getSettingsFiles(bool $import): array{
		return array_filter([
			'cliConfig.inc.php' => WE_CUSTOM_DIR . 'conf/cliConfig.inc.php',
			'we_conf_global.inc.php' => WE_CUSTOM_DIR . 'conf/we_conf_global.inc.php',
			'we_conf_language.inc.php' => WE_CUSTOM_DIR . 'conf/we_conf_language.inc.php',
			'we_active_integrated_modules.inc.php' => WE_CUSTOM_DIR . 'conf/we_active_integrated_modules.inc.php',
			'agency.inc.php' => ($import || file_exists(WE_CUSTOM_DIR . 'agency.inc.php') ?
			WEBEDITION_DIR . 'agency.inc.php' :
			'')
		]);
	}

	static function getRealTableName(string $table): string{
		$settings = we_backup_settings::i();
		'@phan-var we_backup_settings $settings';

		$table = stripTblPrefix(strtolower($table));
		$match = [];
		if(preg_match('|tblobject_([0-9]*)$|', $table, $match)){
			return (isset($settings->tables['tblobject_']) ?
				$settings->tables['tblobject_'] . $match[1] :
				'');
		}

		return ( $settings->tables[$table] ?? '');
	}

	static function getDefaultTableName(string $table): string{
		$settings = we_backup_settings::i();
		'@phan-var we_backup_settings $settings';
		if(($no = we_contents_object::isObjectXTable($table))){
			if(isset($settings->tables['tblobject_'])){
				$max = fI('SELECT MAX(ID) FROM ' . OBJECT_TABLE);
				if($no <= $max){
					return 'tblobject_' . $no;
				}
			}

			return '';
		}

//$def_table = array_search($table,we_backup_settings::i()['tables']);
		foreach($settings->tables as $key => $value){
			if(strtolower($table) == strtolower($value)){
				return $key;
			}
		}
		return '';
	}

	public static function getDescription(string $table, string $prefix): string{
		switch($table){
			case CONTENT_TABLE:
				return g_l('backup', '[' . $prefix . '][content]');
			case FILE_TABLE:
				return g_l('backup', '[' . $prefix . '][files]');
			case TEMPLATES_TABLE:
				return g_l('backup', '[' . $prefix . '][templates]');
			case HISTORY_TABLE:
				return g_l('backup', '[' . $prefix . '][history_data]');
			case INDEX_TABLE:
				return g_l('backup', '[' . $prefix . '][indexes]');
			case DOC_TYPES_TABLE:
				return g_l('backup', '[' . $prefix . '][doctypes]');
			case USER_TABLE:
				return g_l('backup', '[' . $prefix . '][user_data]');
			case CUSTOMER_TABLE:
				return g_l('backup', '[' . $prefix . '][customer_data]');
			case SHOP_ORDER_TABLE:
			case SHOP_ORDER_DATES_TABLE :
			case SHOP_ORDER_DOCUMENT_TABLE :
			case SHOP_ORDER_ITEM_TABLE:
				return g_l('backup', '[' . $prefix . '][shop_data]');
			case PREFS_TABLE :
				return g_l('backup', '[' . $prefix . '][prefs]');
			default:
				return g_l('backup', '[working]');
		}
	}

	public static function getImportPercent(): float{
		$settings = we_backup_settings::i();
		'@phan-var we_backup_settings $settings';

		if($settings->files_to_delete_count){
			$rest1 = (int) ($settings->files_to_delete_count - count($settings->files_to_delete));
			$rest2 = (int) $settings->files_to_delete_count;
		} else {
			$rest1 = 0;
			$rest2 = 0;
		}

		$percent = round(((float)
			(((int) $settings->offset + $rest1) /
			((int) $settings->offset_end + $rest2))) * 100, 2);

		return max(min($percent, 100), 0);
	}

	public static function getExportPercent(): float{
		$settings = we_backup_settings::i();
		'@phan-var we_backup_settings $settings';
		$all = ((int) $settings->row_count) +
			((int) $settings->extern_files_count);

		$done = ((int) $settings->row_counter) +
			($settings->extern_files_count ?
			((int) $settings->extern_files_count) - count($settings->extern_files) :
			0);

		$percent = round(($done / $all) * 100, ($all > 50000 ? 2 : 1));
		return max(min($percent, 100), 0);
	}

	public static function canImportBinary(int $id, string $path): bool{
		$settings = we_backup_settings::i();
		'@phan-var we_backup_settings $settings';

		if($id){
			return $settings->option(we_backup_settings::OPTION_BACKUP_BINARY);
		}
		static $settingsFiles = [];
		$settingsFiles = $settingsFiles ?: we_backup_util::getSettingsFiles(true);
		$isSetting = in_array($path, $settingsFiles, true);

		if(($settings->isHandled(we_backup_settings::OPTION_BACKUP_SETTINGS) && $isSetting) ||
			($settings->option(we_backup_settings::OPTION_BACKUP_EXTERN) && !$isSetting) ||
			($settings->isHandled(we_backup_settings::OPTION_BACKUP_HOOKS) && str_starts_with($path, WE_CUSTOM_PATH . we_hook_handler::DIR)) ||
			($settings->isHandled(we_backup_settings::OPTION_BACKUP_CUSTOMTAGS) && str_starts_with($path, WE_CUSTOM_PATH . we_weTag_tag::DIR)) ||
			($settings->isHandled(we_backup_settings::OPTION_BACKUP_CUSTOMTAGS) && str_starts_with($path, WE_CUSTOM_PATH . we_weTag_twbase::DIR) )
		){
			return true;
		}

		return false;
	}

	public static function canImportVersion(int $id, string $path): bool{
		$settings = we_backup_settings::i();
		'@phan-var we_backup_settings $settings';
		return (!empty($id) && str_contains(strtolower($path), strtolower(WE_VERSION_DIR)) && $settings->isHandled(we_backup_settings::OPTION_BACKUP_VERSION_BIN));
	}

	/**
	 *
	 * @param array<string,string> $files
	 * @param string $dir
	 * @param bool $useIndex
	 * @return void
	 */
	static function exportFiles(array $files, string $dir = '', bool $useIndex = false): void{
		$class = we_backup_settings::i()->class;
		'@phan-var we_backup_exportBase $class';
		if($class->open()){
			foreach($files as $ind => $file){
				$class->writeBinFile($file, $dir, $useIndex ? $ind : '');
			}
			$class->close();
		}
	}

	static function getNextTable(): string{
		$settings = we_backup_settings::i();
		'@phan-var we_backup_settings $settings';

// get all table names from database
		$tables = $settings->allTables;

		while(++$settings->current_table_id < count($tables)){
// get real table name from database
			$table = $tables[$settings->current_table_id]['table_name'];

			if(self::getDefaultTableName($table) !== ''){
				$settings->current_table = $table;
				return $table;
			}
		}

		$settings->current_table = '';
		return '';
	}

	public static function delLog(): void{
		we_base_file::delete(we_backup_settings::i()->backup_log_file);
	}

	public static function addLog(string $log, bool $writeImmediate = false): void{
		$settings = we_backup_settings::i();
		'@phan-var we_backup_settings $settings';

		$settings->backup_log_data .= '[' . date('d-M-Y H:i:s', time()) . '] ' . $log . "\r\n";
		if($writeImmediate || ++self::$logCnt > 50){
			self::writeLog();
		}
	}

	public static function writeLog(): void{
		$settings = we_backup_settings::i();
		'@phan-var we_backup_settings $settings';
		if(empty($settings->backup_log_data)){
			return;
		}
		if($settings->backup_log){
			we_base_file::save($settings->backup_log_file, $settings->backup_log_data, 'ab');
		}
		$settings->backup_log_data = '';
		self::$logCnt = 0;
	}

	public static function getFormat(string $file, int $offset = 0): string{
		if(we_base_file::getCompression($file) === 'zip'){
			$info = we_base_file::getZipFileEntryContent($file, we_backup_exportZIP::INFO_FILE);
			if(!$info){
				return 'unknown';
			}
			$info = we_unserialize($info);
			return is_array($info) && !empty($info['Type']) ? $info['Type'] : 'unknown';
		}
		$part = we_base_file::loadPart($file, $offset, 512, we_base_file::isCompressed($file));

		return (preg_match('|<\?xml |i', $part) ?
			'xml' :
			'unknown');
	}

	public static function getXMLImportType(string $file, bool $iscompr = false, int $end_off = 0): string{
		$try = 0;
		$count = 30;
		$part_len = 16384;
		$part_skip_len = 204800;

		if($end_off == 0){
			$end_off = self::getEndOffset($file, $iscompr);
		}

		$start = $end_off - $part_len;
		try{
			$part = we_base_file::loadPart($file, 0, $part_len, $iscompr);
		} catch (we_exception_file $e){
			return $e->getMessage();
		}

		if(!str_contains($part, we_backup_util::weXmlExImHead)){
			return 'unknown';
		}
		$hasbinary = false;
		while($try < $count){
			if(preg_match('/.*' . we_backup_util::weXmlExImHead . '.*type="backup".*>/', $part)){
				return 'backup';
			} elseif(preg_match('/<(' . we_document_base::BACKUP_XML_TAG . '|' . we_contents_template::BACKUP_XML_TAG . '|' . we_contents_object::BACKUP_XML_TAG . '|' . we_contents_objectFile::BACKUP_XML_TAG . '|' . we_exim_ExIm::INFO_TAG . '|' . we_navigation_navigation::BACKUP_XML_TAG . ')/i', $part)){
				return 'weimport';
			} elseif(str_contains($part, '<' . we_backup_dbTable::BACKUP_XML_TAG)){
				return 'backup';
			} elseif(str_contains($part, '<' . we_backup_binary::BACKUP_XML_TAG)){
				$hasbinary = true;
			} elseif(str_contains($part, '<customer')){
				return 'customer';
			}

			$part = we_base_file::loadPart($file, $start, $part_len, $iscompr);

			$start -= $part_skip_len;

			$try++;
		}

		return ($hasbinary ? 'weimport' : 'unknown');
	}

	public static function getEndOffset(string $filename, bool $iscompressed): int{
		if(!$iscompressed){
			$fh = fopen($filename, 'rb');
			if($fh){
				fseek($fh, 0, SEEK_END);
				$end = ftell($fh);
				fclose($fh);
			} else {
				throw new we_exception_base('File handle not valid');
			}
			if(!$end){
				throw new we_exception_base('Tell unavailable');
			}
			return $end;
		}

		$fh = gzopen($filename, 'rb');
		if($fh){
			while(!gzeof($fh)){
				gzread($fh, 16768);
			}
			$end = gztell($fh);
			gzclose($fh);
		} else {
			throw new we_exception_base('File handle not valid');
		}
		if(!$end){
			throw new we_exception_base('Tell unavailable');
		}
		return $end;
	}

	public static function hasNextTable(): bool{
		$settings = we_backup_settings::i();
		'@phan-var we_backup_settings $settings';

		$current_id = $settings->current_table_id;
		$current_id++;

// get all table names from database
		$tables = $settings->allTables;

		if($current_id < count($tables)){
			$table = $tables[$current_id]['table_name'];
			if(self::getDefaultTableName($table) === ''){
				return false;
			}

			return true;
		}

		return false;
	}

}
