<?php

declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: 66d79e25ceb9 $
 * $Author: Marc Kraemer <marc.kraemer@webedition.org> $
 * $Date: 2021-01-25 22:14 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */

/**
 * Class we_base_errorHandler
 *
 * Provides a error handler for webEdition.
 */
abstract class we_base_errorHandler{

	/** @var bool */
	private static $initialized = false;

	/** @var array<string,bool> */
	private static $handling = [
		'notice' => false,
		'deprecated' => true,
		'warning' => true,
		'error' => true,
		'sql' => true,
		'display' => false,
		'log' => true,
		'send' => false,
		'shutdown' => false,
	];

	private const FILTER_SQL_FUNCTIONS = ['trigger_error'];
	private const FILTER_ERR_FUNCTIONS = [];
	private const E_SQL = -1;
	public const E_JS = -2;
	public const E_EXCEPTION = -3;
	private const IGNORE_GLOBALS = [
		'GLOBALS', '_GET', '_POST', '_REQUEST', '_COOKIE', '_FILES', '_SERVER', '_SESSION',
		'we', 'DB_WE', 'we_doc', 'we_obj', 'WE_MAIN_DOC', 'loader',
		'WE_MAIN_ID', 'WE_MAIN_DOC'
	];

	public static function start(bool $in_webEdition = true): void{
		if(self::$initialized){
			return;
		}
		self::$initialized = true;
		//defined('WE_ERROR_LOG') || define('WE_ERROR_LOG', 1);
		//defined('WE_ERROR_WARNINGS') || define('WE_ERROR_WARNINGS', 1);
		//defined('WE_ERROR_DEPRECATED') || define('WE_ERROR_DEPRECATED', 1);
		//defined('WE_ERROR_HANDLER') || define('WE_ERROR_HANDLER', 1);

		self::$handling['notice'] = defined('WE_ERROR_NOTICES') ? ((bool) WE_ERROR_NOTICES) : false;
		//self::$handling['deprecated'] = defined('WE_ERROR_DEPRECATED') ? (WE_ERROR_DEPRECATED == 1) : false;
		//self::$handling['warning'] = defined('WE_ERROR_WARNINGS') ? (WE_ERROR_WARNINGS == 1 ) : false;
		//self::$handling['log'] = defined('WE_ERROR_LOG') ? (WE_ERROR_LOG == 1 ) : true;
		self::$handling['send'] = (defined('WE_ERROR_MAIL') && defined('WE_ERROR_MAIL_ADDRESS')) ? (WE_ERROR_MAIL == 1) : false;

		// Get way of how to show errors
		if($in_webEdition){
			ini_set('display_errors', '0');
			//we want all errors inside WE
			//we_error_setHandleAll();
		} else {
			self::$handling['display'] = defined('WE_ERROR_SHOW') ? ((bool) WE_ERROR_SHOW ) : true;
		}

		/* if((defined('WE_ERROR_HANDLER') && WE_ERROR_HANDLER) ){ */
		$error_level = 0 +
				(self::$handling['deprecated'] ? E_DEPRECATED | E_USER_DEPRECATED | E_STRICT : 0) +
				(self::$handling['notice'] ? E_NOTICE | E_USER_NOTICE : 0) +
				(self::$handling['warning'] ? E_WARNING | E_CORE_WARNING | E_COMPILE_WARNING | E_USER_WARNING : 0) +
				(self::$handling['error'] ? E_ERROR | E_PARSE | E_CORE_ERROR | E_COMPILE_ERROR | E_USER_ERROR | E_RECOVERABLE_ERROR : 0);
		error_reporting($error_level);
		//ini_set('error_reporting', ''.$error_level);

		set_error_handler('we_base_errorHandler::errorHandler', $error_level);
		register_shutdown_function('we_base_errorHandler::shutdownHandler');
		set_exception_handler('we_base_errorHandler::exceptionHandler');
	}

	/**
	 *
	 * @param array $skip
	 * @return array{string,string,string,int}
	 */
	public static function getBacktrace(array $skip = []): array{
		$detailedError = $caller = $file = $line = '';
		$backtrace = debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS);
		$cnt = 0;
		//$found = false;
		//error handler called directly caused by an error
		foreach($backtrace as $no => $arr){
			if($arr['function'] === 't_e'){
				unset($backtrace[$no - 1]);
				//$found = true;

				break;
			}
		}
		/* if(!$found){
		  $pos = array_search('error_handler', $skip);
		  unset($skip[$pos]);
		  } */

		foreach($backtrace as $arr){
			//NOTE: error_handler holds line no & filename of the callee if not called by t_e
			if((!empty($arr['class']) && $arr['class'] == __CLASS__ && $arr['function'] != 'errorHandler') || in_array($arr['function'], $skip, true)){
				continue;
			}
			if($cnt === 0 || $cnt === 1 && $arr['function'] === 't_e'){ //this is the caller
				$caller = $arr['function'];
				$file = (isset($arr['file']) ? self::replaceDocRoot($arr['file']) : '');
				$line = $arr['line'] ?? 0;
				$detailedError = '';
			}
			$detailedError .= '#' . ($cnt++) . ' ' .
					(empty($arr['class']) ? '' : $arr['class'] . $arr['type']) .
					$arr['function'] .
					(empty($arr['args']) ? '' : '(' . implode(',', $arr['args']) . ')') .
					' called at [' . (isset($arr['file']) ? self::replaceDocRoot($arr['file'], '') : '') . ':' . ($arr['line'] ?? '') . "]\n";
		}
		return [$detailedError, $caller, $file, (int) $line];
	}

	public static function setHandleAll(): void{
		self::$handling = [
			'notice' => true,
			'deprecated' => true,
			'warning' => true,
			'error' => true,
			'sql' => true,
			'display' => false,
			'log' => true,
			'send' => (defined('WE_ERROR_MAIL') && defined('WE_ERROR_MAIL_ADDRESS')) ? (WE_ERROR_MAIL == 1) : false,
			'shutdown' => false,
		];
	}

	public static function disable(): void{
		ini_set('error_reporting', '0');
		error_reporting(0);
		set_error_handler(null);
	}

	public static function setHandling(string $type, bool $value): void{
		self::$handling[$type] = $value;
	}

	public static function getHandling(string $type): bool{
		return self::$handling[$type];
	}

	private static function translateType(int $type): string{
		switch($type){
			case E_RECOVERABLE_ERROR:
			case E_ERROR:
				return 'Error';
			case E_WARNING:
				return 'Warning';
			case E_PARSE:
				return 'Parse error';
			case E_NOTICE:
				return 'Notice';
			case E_CORE_ERROR:
				return 'Core error';
			case E_CORE_WARNING:
				return 'Core warning';
			case E_COMPILE_ERROR:
				return 'Compile error';
			case E_COMPILE_WARNING:
				return 'Compile warning';
			case E_USER_ERROR:
				return 'User error';
			case E_USER_WARNING:
				return 'User warning';
			case E_USER_NOTICE:
				return 'User notice';
			case E_DEPRECATED:
				return 'Deprecated notice';
			case E_STRICT:
				return 'Strict Error';
			case E_USER_DEPRECATED:
				return 'User deprecated notice';
			case self::E_SQL:
				return 'SQL Error';
			case self::E_JS:
				return 'JS Error';
			case self::E_EXCEPTION:
				return 'Exception';
			default:
				return 'unknown Error(' . $type . ')';
		}
	}

	private static function showDevice(int $type, string $message, string $file, int $line, bool $skip = false, string $bt = ''): void{
		// Display error?
		if(self::$handling['display']){
			self::display($type, $message, $file, $line, $skip, $bt);
		}

		// Log error?
		$insertID = (self::$handling['log'] ? self::log($type, $message, $file, $line, $skip, $bt) : 0);

		// Mail error?
		if(!empty(self::$handling['send'])){
			self::mail($type, $message, $file, $line, $skip, $bt, $insertID);
		}
	}

	private static function mail(int $type, string $message, string $file, int $line, bool $skipBT = false, string $bt = '', int $insertID = 0): void{
		static $max = 15;
		if(--$max < 0){
			//don't mail more than this
			return;
		}
		$detailedError = $caller = '-';
		if(!$skipBT){
			[$detailedError, $caller, $file, $line] = self::getBacktrace(($type == self::E_SQL ? self::FILTER_SQL_FUNCTIONS : self::FILTER_ERR_FUNCTIONS));
		} else if($bt){
			$detailedError = $bt;
			//$caller= $file, $line) = $skipBT;
		}

		$ttype = self::translateType($type);

		// Build the error table
		$detailedError = "An $ttype occurred while executing a script in webEdition.\n\n\n" .
				($insertID && function_exists('getServerUrl') ?
				we_http_util::getServerUrl() . WEBEDITION_DIR . 'we_cmd.php?we_cmd[0]=we_gui_showerrorlog&function=pos&ID=' . $insertID . "\n\n" : '') .
// Domain
				'webEdition address: ' . $_SERVER['SERVER_NAME'] . ',

URI: ' . ($_SERVER['REQUEST_URI'] ?? '-none set-') . '
Host: ' . gethostname() . '
Error type: ' . $ttype . "
Error message: " . self::replaceDocRoot($message, 'SECURITY_REPL_DOC_ROOT/') . '
Script name: ' . self::replaceDocRoot($file, 'SECURITY_REPL_DOC_ROOT/') . '
Line number: ' . $line . '
Caller: ' . $caller . '
Backtrace: ' . $detailedError;

		$show = !defined('WE_ERROR_MAIL_ADDRESS') || !mail(WE_ERROR_MAIL_ADDRESS, mb_encode_mimeheader($ttype . ': ' . $_SERVER['SERVER_NAME'] . '(webEdition)', 'UTF-8', 'Q') ?: '',
						strtr(quoted_printable_encode($detailedError), ['=0A' => "\n"]),
						implode("\r\n", [
					'From' => 'From: ' . WE_DEFAULT_EMAIL,
					'Content-Type' => 'Content-Type: text/plain; charset=UTF-8',
					'Content-Transfer-Encoding' => 'Content-Transfer-Encoding: Quoted-Printable'
		]));
		// Log the error
		if($show && ($type & (E_ERROR | E_CORE_ERROR | E_COMPILE_ERROR | E_USER_ERROR)) > 0){
			echo 'Cannot log error! Could not send e-mail due to no known recipient: <pre>' . $detailedError . '</pre>';
		}
	}

	private static function display(int $type, string $message, string $file, int $line, bool $skipBT = false, string $bt = ''): void{
		$detailedError = '-';
		if(!$skipBT){
			[$detailedError, /* $caller */, $file, $line] = self::getBacktrace(($type == self::E_SQL ? ['error_showDevice', 'trigger_error', 'error_handler', 'getBacktrace',
						'display_error_message'] :
							['error_showDevice', 'error_handler', 'getBacktrace', 'display_error_message']));
		} else if($bt){
			$detailedError = $bt;
		}

		// Build the error table
		echo '<br/><table bgcolor="#FFFFFF" cellpadding="4" style="text-align:center;border: 1px solid #265da6;width:95%"><colgroup><col sytle="width:10%"/><col style="width:90%" /></colgroup>
		<tr bgcolor="#f7f7f7" style="vertical-align:top">
			<td colspan="2" style="border-bottom: 1px solid #265da6;">An error occurred while executing this script.</td>
		</tr>
	<tr style="vertical-align:top">
		<td style="white-space:nowrap;border-bottom: 1px solid #265da6; border-right: 1px solid #265da6;font-weight:bold;">Error type:</td>
		<td style="border-bottom: 1px solid #265da6;"><i>' . self::translateType($type) . '</i></td>
	</tr>
	<tr style="vertical-align:top">
			<td style="white-space:nowrap;border-bottom: 1px solid #265da6; border-right: 1px solid #265da6;font-weight:bold;">Error message:</td>
			<td style="border-bottom: 1px solid #265da6;"><i><pre>' . self::replaceDocRoot($message, '') . '</pre></i></td>
	</tr>
	<tr style="vertical-align:top">
			<td style="white-space:nowrap;border-bottom: 1px solid #265da6; border-right: 1px solid #265da6;font-weight:bold;">Script name:</td>
			<td style="border-bottom: 1px solid #265da6;"><i>' . self::replaceDocRoot($file, '') . '</i></td>
	</tr>
	<tr style="vertical-align:top">
			<td style="white-space:nowrap;border-bottom: 1px solid #265da6; border-right: 1px solid #265da6;font-weight:bold;">Line number:</td>
			<td style="border-bottom: 1px solid #265da6;"><i>' . $line . '</i></td>
	</tr>
	<tr style="vertical-align:top">
			<td style="white-space:nowrap;border-right: 1px solid #265da6;font-weight:bold;">Backtrace</td>
			<td>' . str_replace(["\r", "\n"], '', nl2br($detailedError)) . '</td>
	</tr>
	</table><br/>';
	}

	public static function log(int $type, string $message, string $file, int $line, bool $skipBT = false, string $bt = ''): int{
		static $max = 500;
		if(--$max < 0){
			//don't log more messages per request
			return 0;
		}
		$detailedError = $caller = '-';
		if(!$skipBT){
			[$detailedError, $caller, $file, $line] = self::getBacktrace(($type == self::E_SQL ? ['error_showDevice', 'trigger_error', 'error_handler', 'getBacktrace',
						'log_error_message'] :
							['error_showDevice', 'error_handler', 'getBacktrace', 'log_error_message']));
		} else if($bt){
			$detailedError = $bt;
		}

		$id = 0;
		// Log the error
		if(defined('DB_HOST') && defined('DB_USER') && defined('DB_PASSWORD') && defined('DB_DATABASE')){
			$logVars = ['Request', 'Session', 'Server'];
			$_SERVER['SERVER_HOST'] = gethostname();
			$tbl = defined('ERROR_LOG_TABLE') ? ERROR_LOG_TABLE : TBL_PREFIX . 'tblErrorLog';
			$query = 'INSERT INTO ' . $tbl . ' SET ' . we_database_base::arraySetter([
						'Type' => self::translateType($type),
						'Function' => mb_substr($caller, 0, 250),
						'File' => mb_substr(self::replaceDocRoot($file, 'SECURITY_REPL_DOC_ROOT/'), 0, 250),
						'Line' => (int) $line,
						'Text' => mb_substr(preg_replace('_[^[:print:]\n\t ]_u', '.', $message) ?: $message, 0, 60000),
						'Backtrace' => $detailedError,
						'Request' => '',
						'Session' => '',
						'Global' => '',
						'Server' => ''
			]);
			if(isset($GLOBALS['DB_WE'])){
				$db = new DB_WE();
				if(!$db->query($query)){
					self::mail($type, 'Cannot log error! Query failed: ' . $db->Error . "\nError message: " . $message, $file, (int) $line, $skipBT);
				} else {
					$id = $db->getInsertId();
					foreach($logVars as $var){
						$db->query('UPDATE IGNORE ' . $tbl . ' SET ' . self::getDBVariableMax($var) . ' WHERE ID=' . $id);
					}
				}
			} else {
				$hostPort = explode(':', DB_HOST);
				$Host = $hostPort[0];
				$Port = (int) (($hostPort[1] ?? ini_get('mysqli.default_port')) ?: 3306);

				$link = mysqli_connect($Host, DB_USER, DB_PASSWORD, DB_DATABASE, $Port);
				if(!$link){
					exit('Cannot log error! Could not connect ' . mysqli_connect_error());
				}
				if(mysqli_query($link, $query) === false){
					self::mail($type, 'Cannot log error! Query failed: ' . $message, $file, $line, $skipBT);
					//exit('Cannot log error! Query failed: ' . mysql_error());
				} else {
					$id = (int) mysqli_insert_id($link);
					foreach($logVars as $var){
						mysqli_query($link, 'UPDATE ' . $tbl . ' SET ' . self::getDBVariableMax($var) . ' WHERE ID=' . $id);
					}
				}
				mysqli_close($link);
			}
		} else {
			self::mail($type, 'Cannot log error! Database connection not known: ' . $message, $file, $line, $skipBT);
			//exit('Cannot log error! Database connection not known.');
		}
		return $id;
	}

	private static function getDBVariableMax(string $var): string{
		static $max = 65000; //max lenght of text-col in mysql - this is enough debug-data, leave some space...
		$ret = '';
		switch($var){
			case 'Request':
				$ret = (isset($_REQUEST) ? self::prettyPrintArray(print_r(array_diff_key($_REQUEST, [
							'user' => '',
							'username' => '',
							'pass' => '',
							'password' => '',
							's' => '',
							'WE_LOGIN_password' => '',
							'WE_LOGIN_username' => '',
							'Password' => '',
							'Password2' => '',
										]), true), 1) : ' - ');
				break;
			case 'Session':
				if(!isset($_SESSION)){
					$ret = ' - ';
					break;
				}
				$ret = '';
				//FIXME: clone will be reduced to unsetting weS+webuser if all vars have moved
				if(!empty($_SESSION['webuser']['registered'])){
					$ret .= 'Customer: ' .
							self::prettyPrintArray(print_r(['ID' => $_SESSION['webuser']['ID'], 'Username' => $_SESSION['webuser']['Username'] . ' (' . $_SESSION['webuser']['Forename'] . ' ' . $_SESSION['webuser']['Surname'] . ')'], true)) .
							"----------------------------------------\n";
				}
				if(($uid = we_users_user::getID())){
					$ret .= 'User: ' .
							self::prettyPrintArray(print_r(['ID' => $uid, 'Username' => ($_SESSION['user']['Username'] ?? '') . ' (' . ($_SESSION['user']['RealName'] ?? '') . ')'], true)) .
							"----------------------------------------\n";
				}

				if(isset($_SESSION['weS'])){
					$ret .= "Internal data:\n" .
							self::prettyPrintArray(print_r(array_diff_key($_SESSION['weS'], [
								'versions' => '',
								'prefs' => '',
								'we_data' => '',
								'perms' => '',
								'webuser' => ''
											]), true), 1) .
							"----------------------------------------\n";
				}

				if(isset($_SESSION['perms'])){
					$ret .= "Effective Permissions:\n" . self::prettyPrintArray(print_r(array_filter($_SESSION['perms']), true)) .
							"\n------------------------------------\n";
				}
				$ret .= self::prettyPrintArray(print_r(array_diff_key($_SESSION, [
							'prefs' => '',
							'perms' => '',
							'webuser' => '',
							'weS' => '',
							'translations_version' => '', //adminer
							'translations' => '', //adminer
							'token' => '', //adminer
							'pwds' => '', //adminer
							'db' => '', //adminer
							'dbs' => '', //adminer
										]), true), 1);

				break;
			case 'Global':
				if(!isset($GLOBALS)){
					$ret = ' - ';
					break;
				}
				$clone = [];
				foreach($GLOBALS as $key => $val){
					if(!in_array($key, self::IGNORE_GLOBALS, true)){
						$clone[] = $val;
					}
				}
				$ret .= print_r($clone, true);

				break;
			case 'Server':
				if(empty($_SERVER)){
					$ret = ' - ';
				} else {
					$s = $_SERVER;
					ksort($s);
					$ret = self::prettyPrintArray(print_r($s, true));
				}
				break;
			default:
				$ret = '';
		}

		if(strlen($ret) > $max){
			$ret = substr($ret, 0, $max) . "\n[...]";
		}
		return $var . '="' . we_database_base::escape_sql($ret) . '"';
	}

	private static function prettyPrintArray(string $var, int $unindent = 0): string{
		return preg_replace(['|Array\n\(|', '|\n\)$|', '|\n(    )' . ($unindent ? '{' . $unindent . '}' : '+' ) . '|', '|\n(    )|'], ['', '', "\n", "\n\t"], $var);
	}

	public static function errorHandler(int $type, string $message, string $file = '', int $line = 0, array $context = []): bool{//@phan-suppress-current-line PhanUnusedPublicNoOverrideMethodParameter
		// Don't respond to the error if it was suppressed with a '@'
		if(error_reporting() == 0){
			return true;
		}
		if(preg_match('/^MYSQL-(ERROR|WARNING)/', $message)){
			$type = self::E_SQL;
			if(!self::$handling['sql']){
				//sql-handling disabled
				return true;
			}
		}

		switch($type){
			case E_NOTICE:
			case E_USER_NOTICE:
				if(self::$handling['notice']){
					self::showDevice($type, $message, $file, $line);
				}
				break;

			case self::E_SQL:
			case E_WARNING:
			case E_CORE_WARNING:
			case E_COMPILE_WARNING:
			case E_USER_WARNING:
				if(self::$handling['warning']){
					self::showDevice($type, $message, $file, $line);
				}

				break;

			case E_ERROR:
			case E_PARSE:
			case E_CORE_ERROR:
			case E_COMPILE_ERROR:
			case E_USER_ERROR:
			case E_RECOVERABLE_ERROR:
				if(self::$handling['error']){
					self::showDevice($type, $message, $file, $line, true);
					//try to close the session!
					if(session_status() === PHP_SESSION_ACTIVE){
						session_write_close();
					}
				}

				// Stop execution
				return false;
			case E_DEPRECATED:
			case E_USER_DEPRECATED:
				if(self::$handling['deprecated']){
					self::showDevice($type, $message, $file, $line);
				}
				break;
			default:
		}
		//Error handled
		return true;
	}

	public static function shutdownHandler(): void{
		if(self::$handling['shutdown'] !== false){
			return;
		}
		$error = error_get_last();
		if(is_array($error)){
			switch($error['type']){
				case E_ERROR:
				case E_PARSE:
				case E_CORE_ERROR:
				case E_COMPILE_ERROR:
				case E_USER_ERROR:
				case E_RECOVERABLE_ERROR:
					self::errorHandler($error['type'], $error['message'] . "\n" . print_r($error, true), $error['file'], $error['line'], []);
				default:
			}
		}
	}

	public static function exceptionHandler(Throwable $exception): void{
		$type = self::E_EXCEPTION;
		$message = $exception->getMessage();
		$file = $exception->getFile();
		$line = $exception->getLine();
		//add current line to backtrace, otherwise it is not included in source around
		$bt = '#0 ' . $file . '(' . $line . '): ' . "\n" .
				$exception->getTraceAsString();

		self::showDevice($type, $message, $file, $line, true, $bt);
	}

	private static function replaceDocRoot(string $where, string $with = ''): string{
		return strtr($where, [
			WE_DOCUMENT_ROOT . '/' => $with,
			$_SERVER['DOCUMENT' . '_ROOT'] . '/' => $with,
			$_SERVER['DOCUMENT' . '_ROOT'] => $with
		]);
	}

}
