<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: 9fb905d0c6b4 $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2021-01-15 19:34 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */
abstract class we_base_file{
	public const SZ_HUMAN = 0;
	public const SZ_BYTE = 1;
	public const SZ_KB = 2;
	public const SZ_MB = 3;
	public const ERROR_NO_SUCH_FILE = -1;
	public const ERROR_DUPLICATE_NAME = -2;
	public const ERROR_SAME_PARENT = -4;

	static function load(string $filename, string $flags = 'rb', int $rsize = 8192, bool $iscompressed = false): string{
		if($filename === ''){
			throw new we_exception_file('No filename');
		}
		if(!self::hasURL($filename)){
			$oldFile = $filename;
			$filename = realpath($filename);
			/* if(strpos($filename, WE_DOCUMENT_ROOT) === false){
			  t_e('warning', 'Acess outside document_root forbidden!', $filename);
			  return;
			  } */

			if(empty($filename) || !is_readable($filename) || is_dir($filename)){
				throw new we_exception_file('File ' . $oldFile . ' as real ' . $filename . ' not found/readable');
			}
		}

		$prefix = $iscompressed ? self::getComPrefix('gzip') : 'f';
		$open = $prefix . 'open';
		$read = $prefix . 'read';
		$close = $prefix . 'close';

		$buffer = '';
		if(($fp = @$open($filename, $flags))){
			do{
				$data = $read($fp, $rsize);
				if(strlen($data) == 0){
					break;
				}
				$buffer .= $data;
			} while(true);
			$close($fp);
			return $buffer;
		}
		throw new we_exception_file('File ' . $filename . ' not readable');
	}

	static function loadLine(string $filename, int $offset = 0, int $rsize = 8192, bool $iscompressed = false): string{
		if($filename === '' || self::hasURL($filename) || !is_readable($filename) || is_dir($filename)){
			throw new we_exception_file('File not found/readable');
		}
		$filename = realpath($filename);
		/* if(strpos($filename, WE_DOCUMENT_ROOT) === false){
		  t_e('warning', 'Acess outside document_root forbidden!', $filenam);
		  return;
		  } */

		$prefix = $iscompressed ? self::getComPrefix('gzip') : 'f';
		$open = $prefix . 'open';
		$seek = $prefix . 'seek';
		$read = $prefix . 'read';
		$close = $prefix . 'close';

		if(($fp = $open($filename, 'rb'))){
			if($seek($fp, $offset, SEEK_SET) == 0){
				$buffer = $read($fp, $rsize);
				$close($fp);
				return $buffer;
			} else {
				$close($fp);
			}
		}
		throw new we_exception_file('File not readable');
	}

	/**
	 *
	 * @param string $filename
	 * @param int $from
	 * @param int $to
	 * @param bool $iscompressed
	 * @return string[]
	 * @throws we_exception_file
	 */
	static function loadLines(string $filename, int $from, int $to, bool $iscompressed = false): array{
		if($filename === '' || self::hasURL($filename) || !is_readable($filename) || is_dir($filename)){
			throw new we_exception_file('File not found/readable' . $filename);
		}
		$filename = realpath($filename);
		/* if(strpos($filename, WE_DOCUMENT_ROOT) === false){
		  t_e('warning', 'Acess outside document_root forbidden!', $filenam);
		  return;
		  } */

		$prefix = $iscompressed ? self::getComPrefix('gzip') : 'f';
		$open = $prefix . 'open';
		$gets = $prefix . 'gets';
		$close = $prefix . 'close';

		$lines = [];
		$line = 0;
		if(($fp = $open($filename, 'rb'))){
			while((($buffer = $gets($fp, 4096)) !== false) && ++$line < $to){
				if($line >= $from){
					$lines[$line] = $buffer;
				}
			}
			$close($fp);
		}
		return $lines;
	}

	static function loadPart(string $filename, int $offset = 0, int $rsize = 8192, bool $iscompressed = false): string{
		if($filename === '' || self::hasURL($filename) || !is_readable($filename) || is_dir($filename)){
			throw new we_exception_file('File not found/readable');
		}
		$realFile = realpath($filename);
		/* if(strpos($filename, WE_DOCUMENT_ROOT) === false){
		  t_e('warning', 'Acess outside document_root forbidden!', $filename);
		  return;
		  } */

		$prefix = $iscompressed ? self::getComPrefix('gzip') : 'f';
		$open = $prefix . 'open';
		$seek = $prefix . 'seek';
		$read = $prefix . 'read';
		$close = $prefix . 'close';

		if(($fp = @$open($realFile, 'rb'))){
			if($seek($fp, $offset, SEEK_SET) == 0){
				$buffer = $read($fp, $rsize);
				$close($fp);
				return $buffer;
			}
			$close($fp);
		}
		throw new we_exception_file('invalid position');
	}

	static function save(string $filename, string $content, string $flags = 'wb', string $compression = ''): bool{
		if(empty($filename) || ($url = self::hasURL($filename)) || ($dir = is_dir($filename)) || ($notwrite = (file_exists($filename) && !is_writable($filename)))){
			t_e('error writing file', $filename, $url ?? '', $dir ?? '', $notwrite ?? '', ($notwrite ?? false) ? stat($filename) : '');
			throw new we_exception_file('error writing file ' . $filename);
		}
		$prefix = $compression ? self::getComPrefix($compression) : 'f';
		$open = $prefix . 'open';
		$write = $prefix . 'write';
		$close = $prefix . 'close';
		$len = strlen($content);

		if(file_exists($filename) && str_contains($flags, 'w')){// we need this for hardlinked files! If we write directly to the file, all linked contents will get overwritten
			unlink($filename);
		}

		if(($fp = $open($filename, $flags))){
			$written = $write($fp, $content, $len);
			$close($fp);
			//if we write a php file, invalidate cache if used.
			if(str_ends_with($filename, '.php') && function_exists('opcache_invalidate')){
				opcache_invalidate($filename, true);
			}

			return $written === $len;
		}
		t_e('error writing file, unable to open file', $filename);
		throw new we_exception_file('error writing file, unable to open file ' . $filename);
	}

	static function saveTemp(string $content, string $filename = '', int $lifetime = 300): string{
		$filename = TEMP_PATH . ($filename ?: self::getUniqueId());
		if(self::save($filename, $content, 'wb')){
			self::insertIntoCleanUp($filename, $lifetime);
			return $filename;
		}
		throw new we_exception_file('unable to save temp file');
	}

	static function delete(string $filename): bool{
		if(!$filename){
			return false;
		}
		if(!self::hasURL($filename) && (is_writable($filename) || is_link($filename))){
			return is_dir($filename) && !is_link($filename) ? self::deleteLocalFolder($filename) : self::deleteLocalFile($filename);
		}
		return false;
	}

	static function hasURL(string $filename): bool{
		$f = strtolower($filename);
		return str_starts_with($f, 'http') || str_starts_with($f, 'ftp');
	}

	static function getUniqueId(bool $md5 = true): string{
		// md5 encrypted hash with the start value microtime(). The function
		// uniqid() prevents from simultanious access, within a microsecond.
		return ($md5 ? md5(uniqid(__FILE__, true)) : str_replace('.', '', uniqid('', true)));
	}

	/**
	 * Function: splitFile
	 *
	 * Description: This function splits a file.
	 */
	static function splitFile(string $filename, string $path, string $pattern = '', int $split_size = 0, string $marker = ''): int{

		if(empty($pattern)){
			$pattern = basename($filename) . '%s';
		}
		$buff = '';
		$fh_temp = 0;
		$fh = fopen($filename, 'rb');
		$num = -1;
		$open_new = true;
		$fsize = 0;

		$marker_size = strlen($marker);

		if($fh){
			while(!@feof($fh)){
				$line = '';
				$findline = false;

				while($findline == false && !@feof($fh)){
					$line .= @fgets($fh, 4096);
					if(str_ends_with($line, "\n")){
						$findline = true;
					}
				}

				if($open_new){
					$num++;
					$filename_tmp = sprintf($path . $pattern, $num);
					$fh_temp = fopen($filename_tmp, 'wb');
					$open_new = false;
				}

				if($fh_temp){
					$buff .= $line;

					if($marker_size){
						$write = ((substr($buff, (0 - ($marker_size + 1))) === $marker . "\n") || (substr($buff, (0 - ($marker_size + 2))) === $marker . "\r\n"));
					} else {
						$write = true;
					}

					if($write){
						$fsize += strlen($buff);
						fwrite($fh_temp, $buff);
						if(($split_size && $fsize > $split_size) || ($marker_size)){
							$open_new = true;
							@fclose($fh_temp);
							$fsize = 0;
						}
						$buff = '';
					}
				} else {
					return -1;
				}
			}
		} else {
			return -1;
		}
		if($fh_temp){
			if($buff){
				fwrite($fh_temp, $buff);
			}
			@fclose($fh_temp);
		}
		@fclose($fh);

		return $num + 1;
	}

	static function mkpath(string $path): bool{
		$path = str_replace('\\', '/', $path);
		return (self::hasURL($path) ?
			false :
			($path ? self::createLocalFolderByPath($path) : false));
	}

	public static function insertIntoCleanUp(string $path, int $date = 300): void{
		$DB_WE = new DB_WE();
		$date = max($date, 300); //make each entry last at least 300 seconds
		$DB_WE->query('INSERT INTO ' . CLEAN_UP_TABLE . ' SET ' . we_database_base::arraySetter([
				'Path' => rtrim(str_replace(WE_DOCUMENT_ROOT, '', $path), '/'),
				'Date' => sql_function('(NOW()+ INTERVAL ' . ((int) $date) . ' SECOND)'),
			]) . ' ON DUPLICATE KEY UPDATE Date=(NOW()+ INTERVAL ' . ((int) $date) . ' SECOND)');
	}

	public static function deleteLocalFile(string $filename): bool{
		return (file_exists($filename) || is_link($filename) ? unlink($filename) : false);
	}

	public static function checkAndMakeFolder(string $path, bool $recursive = false): bool{
		/* if the directory exists, we have nothing to do and then we return true  */
		if((file_exists($path) && is_dir($path)) || (strtolower(WE_DOCUMENT_ROOT) == strtolower(rtrim($path, '/')))){
			return true;
		}

		// if instead of the directory a file exists, we delete the file and create the directory
		if(file_exists($path) && (!is_dir($path))){
			if(!self::deleteLocalFile($path)){
				t_e('Warning', "Could not delete File '" . $path . "'");
			}
		}

		$umask = umask(0);

		// check for directories: create it if we could no write into it:
		if(!mkdir($path, (int) octdec((string) (WE_NEW_FOLDER_MOD)), $recursive)){
			t_e('Warning', "Could not create local Folder (mask " . WE_NEW_FOLDER_MOD . ") at '" . __FUNCTION__ . "': '" . $path . "'");
			umask($umask); // @phan-suppress-current-line PhanPluginUseReturnValueInternalKnown
			return false;
		}
		umask($umask); // @phan-suppress-current-line PhanPluginUseReturnValueInternalKnown
		return true;
	}

	/**
	 * @deprecated since version 7.0
	 * @param string $RootDir
	 * @param string $path
	 * @return boolean
	 */
	public static function createLocalFolder(string $RootDir, string $path = ''): bool{
		return self::createLocalFolderByPath($RootDir . $path);
	}

	public static function createLocalFolderByPath(string $completeDirPath): bool{
		$returnValue = true;

		if(self::checkAndMakeFolder($completeDirPath, true)){
			return $returnValue;
		}

		$cf = [$completeDirPath];

		$parent = str_replace('\\', '/', dirname($completeDirPath));

		while(!self::checkAndMakeFolder($parent)){
			$cf[] = $parent;
			$parent = str_replace('\\', '/', dirname($parent));
		}
		$umask = umask(0);

		foreach(array_reverse($cf) as $dir){
			if(!mkdir($dir, (int) octdec((string) WE_NEW_FOLDER_MOD), true)){
				t_e('Warning', 'Could not create local Folder at "' . __FUNCTION__ . '": "' . $dir . '"');
				$returnValue = false;
			}
		}
		umask($umask); // @phan-suppress-current-line PhanPluginUseReturnValueInternalKnown

		return $returnValue;
	}

	public static function hasGzip(): bool{
		return function_exists('gzopen');
	}

	public static function hasZip(): bool{
		return function_exists('zip_open');
	}

	public static function hasBzip(): bool{
		return function_exists('bzopen');
	}

	public static function hasCompression(string $comp): bool{
		switch($comp){
			case 'gzip':
				return self::hasGzip();
			case 'zip':
				return self::hasZip();
			case 'bzip':
				return self::hasBzip();
			default:
				return false;
		}
	}

	public static function getComPrefix(string $compression): string{
		switch($compression){
			case 'gzip':
				return 'gz';
			case 'zip':
				return 'zip_';
			case 'bzip':
				return 'bz';
			default://leave here since 0 is equivalent to first switch
				return 'f';
		}
	}

	public static function getZExtension(string $compression): string{
		switch($compression){
			case 'gzip':
				return 'gz';
			case 'zip':
				return 'zip';
			case 'bzip':
				return 'bz';
			default:
				return '';
		}
	}

	static function getCompression(string $filename): string{
		$compressions = ['gzip', 'zip', 'bzip'];
		foreach($compressions as $val){
			if(str_ends_with(strtolower($filename), '.' . self::getZExtension($val))){
				return $val;
			}
		}
		return 'none';
	}

	static function compress(string $file, string $compression = 'gzip', string $destination = '', bool $remove = true, string $writemode = 'wb'): string{
		if(!self::hasCompression($compression)){
			throw new we_exception_file('compression not available:' . $compression);
		}

		$zfile = ($destination ?: $file) . '.' . self::getZExtension($compression);

		if(self::isCompressed($file)){
			if($remove){
				rename($file, $zfile);
			} else {
				copy($file, $zfile);
			}
			return $zfile;
		}
		$prefix = self::getComPrefix($compression);
		$open = $prefix . 'open';
		$write = $prefix . 'write';
		$close = $prefix . 'close';

		$fp = @fopen($file, 'rb');
		if($fp){
			$gzfp = $open($zfile, $writemode);
			if($gzfp){
				do{
					$data = fread($fp, 8192);
					$data_size = $data !== false ? strlen($data) : 0;
					if($data_size == 0){
						break;
					}
					$written = $write($gzfp, $data, $data_size);
					if($data_size != $written){
						throw new we_exception_file('unable to write data');
					}
				} while(true);
				$close($gzfp);
			} else {
				fclose($fp);
				throw new we_exception_file('unable to close file');
			}
			fclose($fp);
		} else {
			throw new we_exception_file('unable to open file');
		}
		if($remove){
			self::delete($file);
		}
		return $zfile;
	}

	static function isCompressed(string $file, int $offset = 0): bool{
		if(!($fh = @fopen($file, 'rb'))){
			return false;
		}
		if(fseek($fh, $offset, SEEK_SET) == 0){
			// according to rfc1952 the first two bytes identify the format
			$id1 = ord((string) fgets($fh, 2));
			$id2 = ord((string) fgets($fh, 2));
			fclose($fh);
			return (($id1 == 31) && ($id2 == 139));
		}
		fclose($fh);
		return false;
	}

	/**
	 * @destination string where the link should point to (fullqualified)
	 * @link string	fullqualified linkname
	 */
	public static function makeSymbolicLink(string $destination, string $link): bool{
		//basename+dirname
		$destinationPath = str_replace(WE_DOCUMENT_ROOT, '', rtrim($destination, '/'));
		$linktarget = realpath(is_link($link) ? dirname($link) . '/' . readlink($link) : $destination);

		if(($linktarget == false || $linktarget != realpath($destination))){
			@unlink($link);
		}
		if(!is_link($link)){
			$cnt = substr_count(str_replace(WE_DOCUMENT_ROOT, '', $link), '/') - 1;
			$destination = str_repeat('../', $cnt) . basename($destinationPath);
			if(!symlink($destination, $link)){
				t_e('error in making symlink from ' . $link . ' to ' . $destination);
			}
		}
		return false;
	}

	/**
	 * creates a hard link to the target - if not possible, make a copy
	 * @param string $destination
	 * @param string $link
	 * @return boolean
	 */
	public static function makeHardLink(string $destination, string $link): bool{
		if(is_file($link)){
			unlink($link);
		}
		if(!is_file($destination)){
			t_e('destination not naming a file', $destination, $link);
			return false;
		}
		if(@link($destination, $link)){
			return true;
		}
		return self::copyFile($destination, $link);
	}

	/**
	 *
	 * @param string $id
	 * @return resource|false
	 */
	static function lock(string $id){
		$path = TEMP_PATH . $id . '.lck';
		$fp = fopen($path, 'c');

		//we can't cleanup file, since another instance might already access this lockfile
		if($fp && flock($fp, LOCK_EX)){
			self::insertIntoCleanUp(TEMP_DIR . $id . '.lck');
			return $fp;
		}
		return false;
	}

	/**
	 *
	 * @param resource $fp
	 * @return void
	 */
	static function unlock(/* resource */$fp): void{
		if(!$fp){
			return;
		}
		flock($fp, LOCK_UN);
		fclose($fp);
	}

	static function getHumanFileSize(int $filesize, int $type = self::SZ_HUMAN): string{
		switch($type){
			case self::SZ_BYTE:
				return $filesize . ' Byte';
			case self::SZ_KB:
				return round($filesize / 1024, 1) . ' kB';
			case self::SZ_MB:
				return round($filesize / (1024 * 1024), 1) . ' MB';
			default:
			case self::SZ_HUMAN:
				if($filesize >= (1024 * 1024)){
					return round($filesize / (1024 * 1024), 1) . ' MB';
				}
				if($filesize >= 1024){
					return round($filesize / 1024, 1) . ' kB';
				}
				return $filesize . ' Byte';
		}
	}

	public static function isWeFile(int $id, string $table = FILE_TABLE): bool{
		if($id === 0){
			return true;
		}
		return (fB('SELECT 1 FROM ' . $table . ' WHERE ID=' . $id));
	}

	public static function cleanTempFiles(bool $cleanSessFiles = false): void{
		$db = DB_WE::$inst;
		$del = [];
		$files = $db->getAllq('SELECT Path FROM ' . CLEAN_UP_TABLE . ' WHERE Date<=NOW() LIMIT 100', true);
		foreach($files as $file){
			if(is_dir(WE_DOCUMENT_ROOT . $file)){
				self::deleteLocalFolder(WE_DOCUMENT_ROOT . $file, true);
			} else {
				self::deleteLocalFile(WE_DOCUMENT_ROOT . $file);
			}
			$del[] = $file;
			if(count($del) > 10){
				$db->query('DELETE FROM ' . CLEAN_UP_TABLE . ' WHERE Path IN ("' . implode('","', $del) . '")');
				$del = [];
			}
		}
		if($cleanSessFiles){
			$sessID = session_id();
			$files = $db->getAllq('SELECT Path FROM ' . CLEAN_UP_TABLE . ' WHERE Path LIKE "%' . DB_WE::$inst->escape($sessID) . '%"', true);
			foreach($files as $file){
				self::deleteLocalFile(WE_DOCUMENT_ROOT . $file);
				$del[] = $file;
			}
		}

		if($del){
			$db->query('DELETE FROM ' . CLEAN_UP_TABLE . ' WHERE Path IN ("' . implode('","', $del) . '")');
		}
	}

	public static function cleanWECache(): void{
		if(defined('WE_VERSION_UPDATE')){
			if(!is_writeable(WE_CACHE_PATH)){
				t_e('cachedir ' . WE_CACHE_PATH . ' is not writeable expect errors & undefined behaviour');
				return;
			}
			we_cache_file::clean();
		}
	}

	public static function we_filenameNotValid(string $filename): bool{
		return (str_starts_with($filename, '..')) || preg_match('![^a-zA-Z0-9~:._\-\[\]\(\)]!', $filename);
	}

	public static function clearPath(string $path): string{
		return preg_replace('#/+#', '/', str_replace('\\', '/', $path));
	}

	public static function replaceInFile(string $string1, string $string2, string $file): void{
		self::save($file, preg_replace('/' . preg_quote($string1, '/') . '/i', $string2, self::load($file, 'r')), 'w');
	}

	public static function deleteLocalFolder(string $filename, bool $delAll = false, bool $withFolder = true, array $skip = [], int $onlyOlder = 0): bool{
		if(!file_exists($filename)){
			return false;
		}
		if($delAll){
			$foo = rtrim($filename, '/') . '/';
			$d = dir($filename);
			while(false !== ($entry = $d->read())){
				switch($entry){
					case '.':
					case '..':
						break;
					case '.htaccess':
						if(!$withFolder){
							break;
						}
					//falls through
					default:
						if(in_array($entry, $skip, true)){//skip this file
							continue 2;
						}
						$path = $foo . $entry;
						if(is_dir($path)){
							//folder mtime is updated when a file is created
							if(!$onlyOlder || filemtime($path) < $onlyOlder){
								self::deleteLocalFolder($path, true, true, [], $onlyOlder);
							}
						} else {
							if(!$onlyOlder || filemtime($path) < $onlyOlder){
								self::deleteLocalFile($path);
							}
						}
				}
			}
			$d->close();
		}
		return $withFolder ? rmdir($filename) : true;
	}

	/**
	 * copy a file
	 * due to windows limitations, the file has to be copied and the old file deleted afterwards.
	 * if $new exists already, windows will not rename the file $old
	 */
	public static function copyFile(string $old, string $new): bool{
		return (@copy($old, $new));
	}

	/**
	 * move/rename a file
	 * due to windows limitations, the file has to be copied and the old file deleted afterwards.
	 * if $new exists already, windows will not rename the file $old
	 */
	public static function moveFile(string $old, string $new): bool{
		if(!@rename($old, $new)){
			if(!copy($old, $new)){
				return false;
			}
			unlink($old);
		}
		return true;
	}

	/**
	 * returns array of directory IDs of all directories which are located inside $folderID (recursive)
	 *
	 * @return int[]
	 * @param int[] $folderIDs
	 */
	public static function getFoldersInFolder(array $folderIDs, string $table = FILE_TABLE): array{
		if(!$folderIDs){
			return [];
		}

		$outArray = array_map('intval', $folderIDs);
		$new = DB_WE::$inst->getAllq('SELECT ID FROM ' . $table . ' WHERE ParentID IN(' . implode(',', $outArray) . ') AND IsFolder=1', true);
		'@phan-var int[] $new';
		return array_merge($outArray, self::getFoldersInFolder($new, $table));
	}

	public static function getZipFileEntryContent(string $file, string $name): string{
		$zip = new ZipArchive();
		if($zip->open($file) !== true){
			throw new we_exception_base('File not found' . $file);
		}
		$ret = $zip->getFromName($name);
		$zip->close();
		return $ret ?: '';
	}

	/**
	 * ensures we have a leading /
	 * @param string $s
	 * @return string
	 */
	public static function addPreSlash(string $s): string{
		return '/' . ltrim($s, '/');
	}

}
