<?php

declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: dfc1b1b26b9f $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2020-12-19 14:26 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */

/**
 * Base Class for all Customer Filters (Model)
 *
 */
abstract class we_customer_abstractFilter{

	/** no filter */
	public const OFF = 0;

	/** all logged in users */
	public const ALL = 1;

	/** only specific customers */
	public const SPECIFIC = 2;

	/** customers with special fields have access */
	public const FILTER = 3;

	/** only not logged in users */
	public const NOT_LOGGED_IN_USERS = 4;

	/** Navigation: filter from document is used */
	public const ACCESS_ON_DOCUMENT = 5;

	/** === */
	public const OP_EQ = 0;

	/** !== */
	public const OP_NEQ = 1;

	/** < */
	public const OP_LESS = 2;

	/** <= */
	public const OP_LEQ = 3;

	/** > */
	public const OP_GREATER = 4;

	/** >= */
	public const OP_GEQ = 5;

	/** ^x */
	public const OP_STARTS_WITH = 6;

	/** x$ */
	public const OP_ENDS_WITH = 7;
	public const OP_CONTAINS = 8;
	public const OP_IN = 9;
	public const OP_NOT_CONTAINS = 10;
	public const OP_NOT_IN = 11;

	/** @var int Mode. Can be OFF, ALL, SPECIFIC, FILTER	 */
	private $mode = self::OFF;

	/** @var int[] Array with customer ids. Only relevant when $_mode is SPECIFIC	 */
	private $specificCustomers = [];

	/** @var int[] Array with customer ids. Only relevant when $_mode is FILTER	 */
	private $blackList = [];

	/** @var int[] Array with customer ids. Only relevant when $_mode is FILTER	 */
	private $whiteList = [];

	/** @var array<int,array{operation:int,field:string,value:string,logic:string}> Array with filter Settings	 */
	private $filter = [];

	/** @var int Id of model (document or object)	 */
	private $modelId = 0;

	/** @var string Table where model is stored in db (eg. FILE_TABLE)	 */
	private $modelTable = '';

	/**
	 * @param integer $mode
	 * @param int[] $specificCustomers
	 * @param int[] $blackList
	 * @param int[] $whiteList
	 * @param array<int,array{operation:int,field:string,value:string,logic:string}> $filter
	 */
	public function __construct(string $modelTable, int $modelId = 0, int $mode = self::OFF, array $specificCustomers = [], array $blackList = [], array $whiteList = [
			], array $filter = [
			]){
		$this->setModelId($modelId);
		$this->setModelTable($modelTable);
		$this->setMode($mode);
		$this->specificCustomers = $specificCustomers;
		$this->blackList = $blackList;
		$this->whiteList = $whiteList;
		$this->filter = $filter;
	}

	/** checks if customer has access with the actual filter object
	 * @return boolean
	 */
	public function customerHasAccess(): bool{
		switch($this->mode){
			case self::OFF:
				return true;
			case self::ALL:
				return self::customerIsLogedIn();
			case self::NOT_LOGGED_IN_USERS:
				return !self::customerIsLogedIn();
			case self::SPECIFIC:
				return self::customerIsLogedIn() && in_array($_SESSION['webuser']['ID'], $this->specificCustomers);
			case self::FILTER:
				return self::customerIsLogedIn() && self::customerHasFilterAccess();
			default:
				return false;
		}
	}

	private static function evalSingleFilter(int $op, string $key, /* mixed */ $value): bool{
		switch($op){
			case self::OP_EQ:
				return $_SESSION['webuser'][$key] == $value;
			case self::OP_NEQ:
				return $_SESSION['webuser'][$key] != $value;
			case self::OP_LESS:
				return $_SESSION['webuser'][$key] < $value;
			case self::OP_LEQ:
				return $_SESSION['webuser'][$key] <= $value;
			case self::OP_GREATER:
				return $_SESSION['webuser'][$key] > $value;
			case self::OP_GEQ:
				return $_SESSION['webuser'][$key] >= $value;
			case self::OP_STARTS_WITH:
				return (str_starts_with((string) $_SESSION['webuser'][$key], $value));
			case self::OP_ENDS_WITH:
				return self::endsWith((string) $_SESSION['webuser'][$key], $value);
			case self::OP_CONTAINS:
				return self::contains((string) $_SESSION['webuser'][$key], $value);
			case self::OP_NOT_CONTAINS:
				return !self::contains((string) $_SESSION['webuser'][$key], $value);
			case self::OP_IN:
				return self::in((string) $_SESSION['webuser'][$key], $value);
			case self::OP_NOT_IN:
				return !self::in((string) $_SESSION['webuser'][$key], $value);
			default:
				t_e('invalid customer filter op: ' . $op);
				return false;
		}
	}

	/** Checks if customer matches $this->filter array
	 * @return boolean
	 */
	private function customerHasFilterAccess(): bool{
		if(in_array($_SESSION['webuser']['ID'], $this->blackList)){
			return false;
		}
		if(empty($this->filter) || in_array($_SESSION['webuser']['ID'], $this->whiteList)){
			return true;
		}
		$hasPermission = false;
		$flag = false;
		$invalidFields = [];
		foreach($this->filter as $filter){
			if(!array_key_exists($filter['field'], $_SESSION['webuser'])){
				$invalidFields[] = (string) $filter['field'];
				continue;
			}
			if($flag && (trim($filter['logic']) === 'AND')){
				$hasPermission = $hasPermission && self::evalSingleFilter((int) $filter['operation'], (string) $filter['field'], (string) $filter['value']);
			} else {
				if($hasPermission){
					$flag = true;
					continue;
				}
				$hasPermission = self::evalSingleFilter((int) $filter['operation'], (string) $filter['field'], (string) $filter['value']);
			}
			$flag = true;
		}

		if($invalidFields){
			t_e('Customerfilter on document/navigation has invalid Parameters, maybe deleted Customer fields: ' . implode(',', array_unique($invalidFields)), $_SESSION['webuser']);
		}

		return $hasPermission;
	}

	public static function evalSingleFilterQuery(int $op, /* mixed */ $key, /* mixed */ $value): string{
		switch($op){
			case self::OP_EQ:
				return '`' . $key . '`="' . $value . '"';
			case self::OP_NEQ:
				return '`' . $key . '`!="' . $value . '"';
			case self::OP_LESS:
				return '`' . $key . '`<"' . $value . '"';
			case self::OP_LEQ:
				return '`' . $key . '`<="' . $value . '"';
			case self::OP_GREATER:
				return '`' . $key . '`>"' . $value . '"';
			case self::OP_GEQ:
				return '`' . $key . '`>="' . $value . '"';
			case self::OP_STARTS_WITH:
				return '`' . $key . '` LIKE "' . $value . '%"';
			case self::OP_ENDS_WITH:
				return '`' . $key . '` LIKE "%' . $value . '"';
			case self::OP_CONTAINS:
				return '`' . $key . '` LIKE "%' . $value . '%"';
			case self::OP_NOT_CONTAINS:
				return '`' . $key . '` NOT LIKE "%' . $value . '%"';
			case self::OP_IN:
				return 'FIND_IN_SET("' . $value . '",`' . $key . '`)';
			case self::OP_NOT_IN:
				return '!FIND_IN_SET("' . $value . '",`' . $key . '`)';
			default:
				t_e('invalid customer filter op: ' . $op);
				return 'FALSE';
		}
	}

	public static function getQueryFromFilter(array $filters): string{
		$ret = '';
		foreach($filters as $filter){
			//FIXME: read webuser table to check for nonexistent fields
			if(!empty($filter['field'])){
				$ret .= ($ret ? ' ' . $filter['logic'] . ' ' : '') . self::evalSingleFilterQuery($filter['operation'], (string) $filter['field'], (string) (empty($filter['value']) ? '' : $filter['value']));
			}
		}
		return $ret ? '(' . $ret . ')' : '';
	}

	/**
	 * Creates and returns the filter array from request
	 *
	 * @static
	 * @return array<int,array{logic:string,field:string,operation:int,value:string}>
	 */
	static function getFilterFromRequest(): array{
		if(we_base_request::string('filterSelect_0', we_base_request::NOT_VALID_S) === we_base_request::NOT_VALID_S){
			return [];
		}
		$filter = [];
		for($count = 0; ($field = we_base_request::string('filterSelect_' . $count)); $count++){
			if(trim(($val = we_base_request::string('filterValue_' . $count)))){
				$filter[] = [
					'logic' => (we_base_request::string('filterLogic_' . $count) === 'OR' ? 'OR' : 'AND'),
					'field' => $field,
					'operation' => we_base_request::int('filterOperation_' . $count),
					'value' => $val
				];
			}
		}
		return $filter;
	}

	/**
	 * Creates and returns the specificCustomers array from
	 *
	 * @static
	 * @return int[]
	 */
	protected static function getSpecificCustomersFromRequest(): array{
		if(!($name = we_base_request::string('specificCustomersEditControl'))){
			return [];
		}
		$customers = [];
		for($i = 0; ($val = we_base_request::int($name . '_variant1_' . $name . '_item' . $i)); $i++){
			$customers[] = $val;
		}
		return $customers;
	}

	/**
	 * Creates and returns the black list array from
	 *
	 * @static
	 * @return int[]
	 */
	protected static function getBlackListFromRequest(): array{
		if(!($name = we_base_request::string('blackListEditControl'))){
			return [];
		}
		$blackList = [];
		for($i = 0; ($val = we_base_request::int($name . '_variant1_' . $name . '_item' . $i)); $i++){
			$blackList[] = $val;
		}
		return $blackList;
	}

	/**
	 * Creates and returns the white list array from request
	 *
	 * @static
	 * @return int[]
	 */
	protected static function getWhiteListFromRequest(): array{
		if(!($name = we_base_request::string('whiteListEditControl'))){
			return [];
		}
		$whiteList = [];
		for($i = 0; ($val = we_base_request::int($name . '_variant1_' . $name . '_item' . $i)); $i++){
			$whiteList[] = $val;
		}
		return $whiteList;
	}

	/**
	 * Checks if $haystack ends with $needle. If so, returns true, otherwise false
	 *
	 * @param string $haystack
	 * @param string $needle
	 * @static
	 * @return boolean
	 */
	static function endsWith(string $haystack, string $needle): bool{
		return (str_ends_with($haystack, $needle));
	}

	/**
	 * Checks if $haystack contains $needle. If so, returns true, otherwise false
	 *
	 * @param string $haystack
	 * @param string $needle
	 * @static
	 * @return boolean
	 */
	static function contains(string $haystack, string $needle): bool{
		return (str_contains($haystack, $needle));
	}

	/**
	 * Checks if $value is one of the CSV Values of $comp
	 *
	 * @param string $value
	 * @param string $comp (CSV)
	 * @static
	 * @return boolean
	 */
	static function in(string $value, string $comp): bool{
		$comp = str_replace('\\,', '__WE_COMMA__', $comp);
		$arr = explode(',', $comp);
		foreach($arr as &$cur){
			$cur = trim(str_replace('__WE_COMMA__', ',', $cur));
		}
		$value = explode(',', $value);
		return count(array_intersect($value, $arr)) > 0;
	}

	/**
	 * Checks if Customer is logged in. Returns true or f alse
	 *
	 * @return boolean
	 */
	public static function customerIsLogedIn(): bool{
		return !empty($_SESSION) && !empty($_SESSION['webuser']) && !empty($_SESSION['webuser']['registered']) && $_SESSION['webuser']['ID'];
	}

	/**
	 * mutator method for $this->mode
	 *
	 * @param integer $mode
	 */
	public function setMode(int $mode): void{
		$this->mode = $mode;
	}

	/**
	 * accessor method for $this->mode
	 *
	 * @return integer
	 */
	public function getMode(): int{
		return $this->mode;
	}

	public function getCustomerFilterFromRequest(int $id): void{
		$this->setModelId($id);
		$this->setMode(($mode = self::getModeFromRequest()));
		$this->specificCustomers = ($mode == self::OFF ? [] : self::getSpecificCustomersFromRequest());
		$this->blackList = ($mode == self::OFF ? [] : self::getBlackListFromRequest());
		$this->whiteList = ($mode == self::OFF ? [] : self::getWhiteListFromRequest());
		$this->filter = ($mode == self::OFF ? [] : self::getFilterFromRequest());
	}

	protected static function getModeFromRequest(): int{
		return we_base_request::int('wecf_mode', self::OFF);
	}

	/**
	 * accessor method for $this->specificCustomers
	 *
	 * @return int[]
	 */
	public function getSpecificCustomers(): array{
		return $this->mode === self::SPECIFIC ? array_filter($this->specificCustomers) : [];
	}

	/**
	 * accessor method for $this->blackList
	 *
	 * @return int[]
	 */
	public function getBlackList(): array{
		return $this->mode === self::FILTER ? array_filter($this->blackList) : [];
	}

	/**
	 * accessor method for $this->whiteList
	 *
	 * @return int[]
	 */
	public function getWhiteList(): array{
		return $this->mode === self::FILTER ? array_filter($this->whiteList) : [];
	}

	/**
	 * mutator method for $this->filter
	 *
	 * @param array<int,array{operation:int,field:string,value:string,logic:string}> $filter
	 */
	public function setFilter(array $filter): void{
		$this->filter = $filter;
	}

	/**
	 * accessor method for $this->filter
	 *
	 * @return array<int,array{operation:int,field:string,value:string,logic:string}>
	 */
	public function getFilter(): array{
		return$this->mode === self::FILTER ? $this->filter : [];
	}

	/**
	 * Mutator method for $this->modelId
	 *
	 * @param integer $modelId
	 */
	public function setModelId(int $modelId): void{
		$this->modelId = $modelId;
	}

	/**
	 * Accessor method for $this->modelId
	 *
	 * @return integer
	 */
	public function getModelId(): int{
		return $this->modelId;
	}

	/**
	 * Mutator method for $this->modelTable
	 *
	 * @param string $modelTable
	 */
	public function setModelTable(string $modelTable): void{
		$this->modelTable = stripTblPrefix($modelTable);
	}

	/**
	 * Accessor method for $this->modelTable
	 *
	 * @return string
	 */
	public function getModelTable(): string{
		return stripTblPrefix($this->modelTable);
	}

	/**
	 * Call this function, when model is deleted !
	 * this function is called, when model with filter is saved (filters are resaved)
	 */
	protected function delete(): void{
		if(($id = $this->getModelId())){
			self::delByModelID($id, $this->getModelTable());
		}
	}

	public static function delByModelID(int $id, string $table){
		DB_WE::$inst->query('DELETE FROM ' . CUSTOMER_FILTER_TABLE . ' WHERE modelId=' . $id . ' AND modelTable="' . $table . '"');
	}

	/**
	 *
	 * @param array<string,mixed> $set
	 * @return void
	 */
	public function save(array $set = []): void{
		if($this->getMode() === self::OFF){
			$this->delete();
			return;
		}

		DB_WE::$inst->query('REPLACE INTO ' . CUSTOMER_FILTER_TABLE . ' SET ' . we_database_base::arraySetter([
					'modelId' => $this->getModelId(),
					'modelTable' => $this->getModelTable(),
					'mode' => $this->getMode(),
					'specificCustomers' => (($specificCustomers = $this->getSpecificCustomers()) ? implode(',', $specificCustomers) : ''),
					'filter' => (($filter = $this->getFilter()) ? we_serialize($filter) : ''),
					'whiteList' => (($whiteList = $this->getWhiteList()) ? implode(',', $whiteList) : ''),
					'blackList' => (($blackList = $this->getBlackList()) ? implode(',', $blackList) : ''),
						] + $set)
		);
	}

	abstract public static function getFilterByDbHash(array $hash);

	/**
	 * @param integer $id
	 * @param string $table
	 * @return array<string,string|int|float>
	 */
	public static function getFilterHashFromDb(int $id, string $table): array{
		return getHash('SELECT * FROM ' . CUSTOMER_FILTER_TABLE . ' WHERE modelTable="' . DB_WE::$inst->escape(stripTblPrefix($table)) . '" AND modelId=' . $id);
	}

	/**
	 * Call this function, if customer is deleted
	 *
	 * @param integer $webUser
	 */
	public static function deleteWebUser(int $webUser): void{
		if(!$webUser){
			return;
		}
		$db = DB_WE::$inst;
		$db->query('UPDATE ' . CUSTOMER_FILTER_TABLE . ' SET specificCustomers=TRIM(BOTH "," FROM REPLACE(CONCAT(",",specificCustomers,","),",' . $webUser . ',", ",")) WHERE FIND_IN_SET(' . $webUser . ',specificCustomers)');
		$db->query('UPDATE ' . CUSTOMER_FILTER_TABLE . ' SET whiteList=TRIM(BOTH "," FROM REPLACE(CONCAT(",",whiteList,","),",' . $webUser . ',", ",")) WHERE FIND_IN_SET(' . $webUser . ',whiteList)');
		$db->query('UPDATE ' . CUSTOMER_FILTER_TABLE . ' SET blackList=TRIM(BOTH "," FROM REPLACE(CONCAT(",",blackList,","),",' . $webUser . ',", ",")) WHERE FIND_IN_SET(' . $webUser . ',blackList)');
	}

	/**
	 * Deletes all filters for given modelIds of table
	 * call this, when several models are deleted
	 */
	public static function deleteModel(array $modelIds, string $table): void{
		if(!$modelIds){
			return;
		}
		DB_WE::$inst->query('DELETE FROM ' . CUSTOMER_FILTER_TABLE . ' WHERE modelId IN (' . implode(', ', $modelIds) . ') AND modelTable="' . DB_WE::$inst->escape(stripTblPrefix($table)) . '"');
	}

}
