<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: ddafdd2b505d $
 * $Author: Nils Reibold <reibold@digirezi.de> $
 * $Date: 2021-01-07 00:43 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */
abstract class we_dialog_metadata{

	/**
	 * This function returns the HTML code of a dialog.
	 *
	 * @param          string                                  $name
	 * @param          string                                  $title
	 * @param          we_html_multiIconBoxContent             $content
	 * @param          int                                     $expand             (optional)
	 * @param          string                                  $show_text          (optional)
	 * @param          string                                  $hide_text          (optional)
	 * @param          bool                                    $cookie             (optional)
	 * @param          string                                  $JS                 (optional)
	 *
	 * @return         string
	 */
	protected static function create_dialog(string $name, $title, we_html_multiIconBoxContent $content, int $expand = -1, string $show_text = '', string $hide_text = '', bool $cookie = false, string $JS = ''): string{

		// Check, if we need to write some JavaScripts
		return
			($JS === '' ? '' : $JS ) .
			// Return HTML code of dialog
				(new we_html_multiIconBox($content, $name))
				->marginLeft()
				->fold($expand, $cookie, $show_text, $hide_text)
				->headline($title)
				->get();
	}

	/**
	 * This functions saves all options.
	 *
	 * @return         void
	 */
	protected static function save_all_values(): void{
		//SAVE METADATA FIELDS TO DB
		if(!we_base_permission::hasPerm([we_base_perm::WORK_METADATAFIELD_EDIT])){
			return;
		}
		DB_WE::$inst->query('TRUNCATE TABLE ' . METADATA_TABLE);
		DB_WE::$inst->query('TRUNCATE TABLE ' . METAVALUES_TABLE);

		if(empty($_REQUEST['metadataTag']) || !is_array($_REQUEST['metadataTag'])){
			return;
		}
		$tagIDs = [];
		foreach(we_base_request::stringList('metadataTag') as $key => $value){
			DB_WE::$inst->query('INSERT INTO ' . METADATA_TABLE . ' SET ' . we_database_base::arraySetter([
					'tag' => $value,
					'type' => we_base_request::string('metadataType', 'textfield', $key),
					'importFrom' => we_base_request::string('metadataImportFrom', '', $key),
					'mode' => we_base_request::string('metadataMode', 'none', $key),
					'csv' => we_base_request::bool('metadataCsv', false, $key),
					'closed' => we_base_request::bool('metadataClosed', false, $key)
			]));
			$tagIDs[$key] = DB_WE::$inst->getInsertId();
		}
		$proposals = we_base_request::stringList('metadataProposal');
		'@phan-var array<string,array<string,string>> $proposals';
		foreach($proposals as $key => $proposals){
			foreach($proposals as $proposal){
				if($proposal){
					DB_WE::$inst->query('INSERT INTO ' . METAVALUES_TABLE . ' SET ' . we_database_base::arraySetter([
							'tagID' => $tagIDs[$key],
							'value' => $proposal
					]));
				}
			}
		}
	}

	protected static function build_dialog(string $selected_setting = 'ui'): string{
		switch($selected_setting){
			// save dialog:
			case 'save':
				$settings = (new we_html_multiIconBoxContent())
					->add(new we_html_multiIconBoxContentLine('', g_l('metadata', '[save]')));
				return self::create_dialog('', g_l('metadata', '[save_wait]'), $settings);
			// SAVED SUCCESSFULLY DIALOG:
			case 'saved':
				$content = (new we_html_multiIconBoxContent())
					->add(new we_html_multiIconBoxContentLine('', g_l('metadata', '[saved]')));
				// Build dialog element if user has permission
				return self::create_dialog('', g_l('metadata', '[saved_successfully]'), $content);
			// THUMBNAILS
			case 'dialog':
				$defined_fields = we_metadata_metaData::getDefinedMetaDataFields(we_metadata_metaData::ALL_BUT_STANDARD_FIELDS);
				'@phan-var array<string,array{tag:string,type:string,importFrom:string,mode:string,csv:int,closed:int}> $defined_fields';
				$defined_values = we_metadata_metaData::getDefinedMetaValues();

				$metadata_types = [
					'textfield' => 'textfield',
					'textarea' => 'textarea',
					//'wysiwyg' 	=> 'wysiwyg',
					'date' => 'date'
				];

				$metadata_modes = [
					'none' => g_l('metadata', '[mode_none]'),
					'manual' => g_l('metadata', '[mode_manually]'),
					'auto' => g_l('metadata', '[mode_auto]')
				];

				$metadata_fields = ['' => '-- ' . g_l('metadata', '[add]') . ' --', 'Exif' => we_html_tools::OPTGROUP];
				foreach(we_metadata_Exif::usedFields as $key){
					$metadata_fields[$key] = $key;
				}
				$metadata_fields['IPTC'] = we_html_tools::OPTGROUP;
				foreach(we_metadata_IPTC::usedFields as $key){
					$metadata_fields[$key] = $key;
				}

				$i = 0;
				$adv_row = '';

				// add standard meta field Keywords
				$standardFields = [];
				foreach(we_metadata_metaData::STANDARD_FIELDS as $name){
					$field = we_metadata_metaData::getMetaDataField($name) ?:
						['tag' => $name,
						'type' => 'textfield',
						'importFrom' => '',
						'mode' => 'none',
						'csv' => ($name === 'Keywords' ? 1 : 0),
						'closed' => 0,
					];
					$field['tagname'] = g_l('weClass', '[' . $name . ']');
					$field['isStandard'] = true;
					$standardFields[] = $field;
				}

				$defined_fields = array_merge($standardFields, array_values($defined_fields));

				//FIXME: clean this html and use we_html_element
				foreach($defined_fields as $key => $value){
					$value['mode'] = $value['mode'] ?: 'none';

					if(empty($value['isStandard'])){
						$row0 = '<td style="width:210px;"><strong>' . g_l('metadata', '[customfield]') . '</strong></td>
	<td style="width:110px;" colspan="2"><strong>' . g_l('metadata', '[type]') . '</strong></td>';
						$row1 = '<td style="width:210px;padding-right:5px;">' . we_html_tools::textInput('metadataTag[' . $key . ']', $value['tag'], 255, [], "text", '205px') . '</td>
	<td style="width:200px;">' . we_html_tools::select('metadataType[' . $key . ']', $metadata_types, 1, $value['type'], false, ['onchange' => "toggleType(this, " . $key . ")"]) . '</td>
	<td style="text-align:right;width:30px;">' . we_html_button::create_button(we_html_button::TRASH, "javascript:delRow(" . $i . ")") . '</td>';
						$row2 = '<td style="padding-bottom:6px;padding-right:5px;">
		<div class="small">' . htmlspecialchars(g_l('metadata', '[import_from]')) . '</div>' . we_html_tools::textInput('metadataImportFrom[' . $key . ']', $value['importFrom'], 255, [
								], "text", '205px') . '
	</td>
	<td colspan="2" style="padding-bottom:6px;">
		<div class="small">' . htmlspecialchars(g_l('metadata', '[fields]')) . '</div>' .
							we_html_tools::select('add_' . $key, $metadata_fields, 1, "", false, ['style' => "width:98%", 'onchange' => "addFieldToInput(this," . $key . ")"]) . '
	</td>';
					} else {
						$row0 = '<td style="width:210px;"><strong>' . g_l('metadata', '[standardfield]') . '</strong></td>
	<td style="width:110px;" colspan="2"><strong>' . g_l('metadata', '[type]') . '</strong></td>';
						$row1 = '<td style="width:210px;padding-right:5px;">' .
							we_html_tools::textInput('tag[' . $key . ']', $value['tagname'], 255, [], "text", "205px", false, true) . '</td>
	<td style="width:200px;">' .
							we_html_tools::select('type[' . $key . ']', $metadata_types, 1, $value['type'], false, ['disabled' => 'disabled']) . '</td>
	<td style="text-align:right;width:30px;">' .
							we_html_element::hiddens([
								'metadataTag[' . $key . ']' => (string) $value['tag'],
								'metadataType[' . $key . ']' => (string) $value['type']
							]) .
							we_html_button::create_button(we_html_button::TRASH, '', '', '', true) . '</td>';
						$row2 = '<td style="padding-bottom:6px;padding-right:5px;">
		<div class="small">' . htmlspecialchars(g_l('metadata', '[import_from]')) . '</div>' . we_html_tools::textInput('metadataImportFrom[' . $key . ']', $value['importFrom'], 255, [
								], "text", '205px') .  '
	</td>
	<td colspan="2" style="padding-bottom:6px;">
		<div class="small">' . htmlspecialchars(g_l('metadata', '[fields]')) . '</div>' .
							we_html_tools::select('add_' . $key, $metadata_fields, 1, "", false, ['style' => "width:98%", 'onchange' => "addFieldToInput(this," . $key . ")"]) . '
	</td>';
					}

					$adv_row .= '
	<tr id="row_' . $key . '">
		<td>
			<table class=" defaultfont" style="background-color:#f5f5f5; margin-bottom:10px" id="elem_' . $key . '">
				<tr id="metadataRow0_' . $key . '">' . $row0 . '
				</tr>
				<tr id="metadataRow1_' . $key . '">' . $row1 . '
				</tr>
				<tr id="metadataRow2_' . $key . '">' . $row2 . '
				</tr>
				<tr id="metadataRow3_' . $key . '">
					<td style="padding-bottom:1px;padding-right:5px;">
						<div class="small" id="metadataModeDiv0_' . $key . '">' . g_l('metadata', '[proposals]') . '</div><div id="metadataModeDiv1_' . $key . '">' . we_html_tools::select('metadataMode[' . $key . ']', $metadata_modes, 1, $value['mode'], false, [
							($value['type'] === 'textfield' ? '' : 'disabled') => ($value['type'] === 'textfield' ? '' : '1'), 'style' => "width:98%", 'onchange' => "togglePropositionTable(this, " . $key . ");"]) . '</div>
					</td>
					<td colspan="2" style="padding-bottom:1px;">
						<div class="small" id="metadataProposalChecks0_' . $key . '">&nbsp;</div>
						<div id="metadataProposalChecks1_' . $key . '">' .
						we_html_forms::checkboxWithHidden((bool) $value['csv'], 'metadataCsv[' . $key . ']', 'CSV', false, 'defaultfont', '', false, '', we_html_tools::TYPE_NONE, 0, '', '', g_l('metadata', '[csv_desc]')) .
						we_html_forms::checkboxWithHidden((bool) $value['closed'], 'metadataClosed[' . $key . ']', g_l('metadata', '[closedList]'), false, 'defaultfont', '', false, '', we_html_tools::TYPE_NONE, 0, '', '', g_l('metadata', '[closedList_desc]')) .
						'</div>
					</td>
				</tr>
				<tr id="metadataRow4_' . $key . '">
					<td colspan="3" style="padding-bottom:16px;padding-right:5px;">
						<table id="proposalTable_' . $key . '" style="width:100%;border:1px solid gray;display:' . ($value['mode'] === 'none' ? 'none' : 'block') . ';padding-top:8px;">';
					if(isset($defined_values[$value['tag']])){
						$i = 0;
						foreach($defined_values[$value['tag']] as $proposal){
							$adv_row .= '<tr>
										<td style="width:15%"></td>
										<td style="text-align:left">' . we_html_tools::textInput('metadataProposal[' . $key . '][' . $i++ . ']', $proposal, 255, ['disabled' => ($value['mode'] === 'auto' ? 'disabled' : '')], "text", '310px') . '</td>
										<td style="width:25px;">' . we_html_button::create_button(we_html_button::TRASH, "javascript:delProposition(this)") . '</td>
									</tr>';
						}
					} else {
						$adv_row .= '<tr>
									<td style="width:15%"></td>
									<td style="text-align:left">' . we_html_tools::textInput('metadataProposal[' . $key . '][0]', '', 255, ['disabled' => ($value['mode'] === 'auto' ? 'disabled' : '')], "text", '310px') . '</td>
									<td style="width:25px;">' . we_html_button::create_button(we_html_button::TRASH, "javascript:delProposition(this)") . '</td>
								</tr>';
					}
					$adv_row .= '<tr>
								<td style="text-align:right;width:15%"></td>
								<td style="text-align:left">' . we_html_button::create_button(we_html_button::PLUS, "javascript:addProposition(this, " . $key . ");") . '</td>
								<td style="width:25px;"></td>
							</tr>
						</table>
					</td>
				</tr>
			</table>
		</td>
	</tr>';
					$i++;
				}

				$metadataTable = '
	<table class="default" style="width:440px;">
		<tbody id="metadataTable">' .
					$adv_row . '
		</tbody>
	</table>';

				$js = we_html_element::jsScript(JS_DIR . 'edit_metadatafields.js', '', ['id' => 'loadVarEdit_metadatafields', 'data-phpdata' => setDynamicVar([
							'tagInp' => we_html_tools::textInput('metadataTag[__we_new_id__]', "", 255, [], "text", '210px'),
							'importInp' => we_html_tools::textInput('metadataImportFrom[__we_new_id__]', "", 255, [], "text", '210px'),
							'typeSel' => we_html_tools::select('metadataType[__we_new_id__]', $metadata_types, 1, 'textfield', false, ['class' => 'defaultfont', 'onchange' => "toggleType(this, __we_new_id__);"]),
							'fieldSel' => we_html_tools::select('add___we_new_id__', $metadata_fields, 1, '', false, ['class' => 'defaultfont', 'style' => 'width:100%',
								'onchange' => "addFieldToInput(this,__we_new_id__);"]),
							'modeSel' => we_html_tools::select('metadataMode[__we_new_id__]', $metadata_modes, 1, 'none', false, ['class' => "defaultfont", 'style' => 'width:100%',
								'onchange' => "togglePropositionTable(this, __we_new_id__);"]),
							'csvCheck' => we_html_forms::checkboxWithHidden(false, 'metadataCsv[__we_new_id__]', 'CSV'),
							'closedCheck' => we_html_forms::checkboxWithHidden(false, 'metadataClosed[__we_new_id__]', g_l('export', '[finish_progress]')),
							'addPropositionBtn' => we_html_button::create_button(we_html_button::PLUS, "javascript:addProposition(this, __we_new_id__);"),
							'trashButton' => we_html_button::create_button(we_html_button::TRASH, "javascript:delRow(__we_new_id__)"),
							'proposalInp' => we_html_tools::textInput('metadataProposal[__we_meta_id__][__we_prop_id__]', "", 255, [], "text", '310px'),
							'delPropositionBtn' => we_html_button::create_button(we_html_button::TRASH, "javascript:delProposition(this);"),
				])]);

				$hint = we_html_tools::alertAttentionBox(g_l('metadata', '[fields_hint]', false, false), we_html_tools::TYPE_INFO, 440, false, 50);
				$hint2 = we_html_tools::alertAttentionBox(g_l('metadata', '[proposals_hint]'), we_html_tools::TYPE_INFO, 440, false);

				$content = $hint . '<div>&nbsp;</div>' . $hint2 . '<div style="height:20px"></div>' . $metadataTable . we_html_button::create_button(we_html_button::PLUS, "javascript:addRow();");

				$contentFinal = (new we_html_multiIconBoxContent())
					->add(new we_html_multiIconBoxContentLine('', $content));
				// Build dialog element if user has permission
				return self::create_dialog('settings_predefined', g_l('metadata', '[headline]'), $contentFinal, -1, '', '', false, $js);
		}
		return '';
	}

	/**
	 * This functions renders the complete dialog.
	 *
	 * @return         string
	 */
	protected static function render_dialog(): string{
		// Render setting groups
		return we_html_element::div(['id' => 'metadatafields_dialog'], self::build_dialog('dialog')) .
			// Render save screen
			we_html_element::div(['id' => 'metadatafields_save', 'style' => 'display: none;'], self::build_dialog('save'));
	}

	protected static function getMainDialog(we_base_jsCmd $jsCmd = null): string{
		// Check if we need to save settings
		if(we_base_request::bool('save_metadatafields')){
			$name = we_base_request::string('metadatafields_name');
			if((str_contains($name, "'") || str_contains($name, ','))){
				$jsCmd->addMsg(g_l('alert', '[metadatafields_hochkomma]'), we_base_util::WE_MESSAGE_ERROR);
				$jsCmd->addCmd('history.back');
			} else {
				self::save_all_values();

				$jsCmd->addMsg(g_l('metadata', '[saved]'), we_base_util::WE_MESSAGE_NOTICE);
				//$save_javascript->addCmd('close_dialog');
			}

			//return $save_javascript->getCmds() .
			//	we_html_element::div(['class' => 'weDialogBody', 'style' => 'height:100%;width:100%'], self::build_dialog('saved'));
		}
		return
			we_html_element::form(['name' => we_base_constants::WE_FORM, 'method' => 'post', 'action' => $_SERVER['REQUEST_URI']], we_html_element::hidden('save_metadatafields', 'false') . self::render_dialog());
	}

	public static function getDialog(): string{
		$jsCmd = new we_base_jsCmd();
		$content = self::getMainDialog($jsCmd);

		return we_html_tools::getHtmlTop(g_l('metadata', '[headline]'), $jsCmd->getCmds(), we_html_element::body(['class' => 'weDialogBody', 'onload' => "self.focus();"]
					, we_html_element::exIFrame('we_metadatafields', $content, 'position:absolute;top:0;bottom:40px;left:0;right:0;overflow:auto;', 'weDialogBody') .
					we_html_element::div(['class' => 'editfooter', 'style' => 'position:absolute;height:40px;bottom:0;left:0;right:0;overflow: hidden;'], we_html_button::position_yes_no_cancel(we_html_button::create_button(we_html_button::OK, "javascript:we_save();"), "", we_html_button::create_button(we_html_button::CLOSE, "javascript:WE().util.dialogTop(window).doClose();")))
		));
	}

}
