<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: 81167abd4e89 $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2021-01-26 23:42 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */

/**
 * class for handling image documents
 */
class we_document_image extends we_document_binary{
	const ALT_FIELD = '_img_custom_alt';
	const TITLE_FIELD = '_img_custom_title';
	const THUMB_FIELD = '_img_custom_thumb';
	const ELEMENT_KEY = [
		'width' => self::ETYPE_NUMBER,
		'height' => self::ETYPE_NUMBER,
		'filesize' => self::ETYPE_NUMBER,
		'origheight' => self::ETYPE_NUMBER,
		'origwidth' => self::ETYPE_NUMBER,
		'parentid' => self::ETYPE_NUMBER,
		'LinkID' => self::ETYPE_NUMBER,
		'RollOverID' => self::ETYPE_NUMBER,
		'thumbnailid' => self::ETYPE_NUMBER,
		'showcontrol' => self::ETYPE_NUMBER
	];
	private const SVG_EXTENSION = ['.svg', '.svgz'];
	public const NEW_SVG_ACTIVE = false;

	/** @var int */
	private static $imgCnt = 0;

	/** @var ?int[] */
	var $Thumbs = null;

	/** @var string ContentType of the Object */
	public $ContentType = we_base_ContentTypes::IMAGE;

	function __construct(int $id = 0){
		self::$imgCnt++;
		parent::__construct($id);
	}

	/** initializes all data not static in class and not saved by persistent slots/elements.
	 * This is called from constructor and __wakeup on serialized documents
	 */
	protected function initStatic(): void{
		parent::initStatic();
		$this->persistent_slots = array_merge($this->persistent_slots, [
			'Thumbs' => we_base_request::INTLISTA,
		]);

		if(isWE() && !self::isSvg($this->Extension)){
			$this->EditPageNrs = array_merge($this->EditPageNrs, [we_base_constants::WE_EDITPAGE_IMAGEEDIT, we_base_constants::WE_EDITPAGE_THUMBNAILS]);
		}
	}

	/**
	 * saves the data of the document
	 *
	 * @return boolean
	 * @param boolean $resave
	 */
	public function we_save(bool $resave = false, bool $skipHook = false): bool{
		// get original width and height of the image
		try{
			$arr = $this->getOrigSize(true, true);
			$this->setElementI('origwidth', $arr[0] ?? 0, 'attrib');
			$this->setElementI('origheight', $arr[1] ?? 0, 'attrib');
		} catch (we_exception_image $e){
			//t_e($e->getMessage(), $e);
			//e.g. svg's
		}
		$docChanged = $this->DocChanged; // will be reseted in parent::we_save()
		//if focus changed, rebuild thumbs
		if($this->ID && ($focus = $this->getElementS('focus'))){
			$oldFocus = fS('SELECT c.Dat FROM ' . CONTENT_TABLE . ' c WHERE c.DocumentTable="' . stripTblPrefix(FILE_TABLE) . '" AND c.DID=' . $this->ID . ' AND c.nHash=x\'' . md5('focus') . '\'');
			$docChanged = $docChanged || ($focus == '[0,0]' && !$oldFocus ? false : ($oldFocus != $focus));
		}
		if(parent::we_save($resave, $skipHook)){
			$this->unregisterMediaLinks();
			$ret = $this->registerMediaLinks();
			$thumbs = self::isSvg($this->Extension) ? [] : $this->getThumbs(); // IMI SVG
			if($docChanged){
				we_contents_thumbnail::deleteByImageID($this->ID);
			}
			foreach($thumbs as $thumbID){
				$thumbObj = new we_contents_thumbnail();
				$thumbObj->initByThumbID((int) $thumbID, $this->ID, $this->Filename, $this->Path, $this->Extension, $this->getElementI('origwidth'), $this->getElementI('origheight'), $this->getDocument());
				if(($docChanged || !$thumbObj->exists()) && !$thumbObj->getThumb()){
					t_e('Error creating thumbnail for file', $this->Filename . $this->Extension);
				}
			}

			return $ret;
		}

		return false;
	}

	function registerMediaLinks(bool $temp = false, bool $linksReady = false): bool{
		if(($id = $this->getElementI('LinkID'))){
			$this->MediaLinks['Hyperlink:Intern'] = $id;
		}
		if(($id = $this->getElementI('RollOverID'))){
			$this->MediaLinks['Hyperlink:Rollover'] = $id;
		}

		return parent::registerMediaLinks(false, true);
	}

	/**
	 * Calculates the original image size of the image.
	 * Returns an array like the PHP function getimagesize().
	 * If the array is empty the image is not uploaded or an error occured
	 *
	 * @param boolean $calculateNew
	 * @throws we_exception_image
	 * @return array
	 */
	function getOrigSize(bool $calculateNew = false, bool $useOldPath = false): array{
		if(self::isSvg($this->Extension)){
			if(self::NEW_SVG_ACTIVE){//@phan-suppress-current-line PhanImpossibleCondition
				return [$this->getElementI('width'), $this->getElementI('height'), 0, ''];
			}
			return [0, 0, 0, ''];
		}
		if(!$this->DocChanged && $this->ID){
			if($this->getElementI('origwidth') && $this->getElementI('origheight') && !$calculateNew){
				return [$this->getElementI('origwidth'), $this->getElementI('origheight'), 0, ''];
			}
			// we have to calculate the path, because maybe the document was renamed
			//$path = $this->getParentPath() . '/' . $this->Filename . $this->Extension;
			return we_contents_thumbnail::getimagesize(WEBEDITION_PATH . '../' . (($useOldPath && $this->OldPath) ? $this->OldPath : $this->Path));
		}
		if(($tmp = $this->getElementS('data'))){
			return we_contents_thumbnail::getimagesize($tmp);
		}
		return [0, 0, 0, ''];
	}

	/**
	 * Returns an array with the Thumbnail IDs for the image.
	 *
	 * @return int[]
	 */
	function getThumbs(): array{//FIXME what about thumbs for svg's?
		if($this->Thumbs !== null){
			return $this->Thumbs;
		}

		$this->Thumbs = [];
		$thumbs = $this->DB_WE->getAllq('SELECT ID,Width,Height,Options,Format,Text,Quality FROM ' . THUMBNAILS_TABLE);
		'@phan-var array<int,array{ID:int,Text:string,Path:string,Format:string,Height:int,Width:int,Options:string,Quality:int}> $thumbs';
		$thumbObj = new we_contents_thumbnail();

		foreach($thumbs as $thumb){
			$thumbObj->init($thumb['ID'], $thumb['Width'], $thumb['Height'], explode(',', $thumb['Options']), $thumb['Format'], $thumb['Text'], $this->ID, $this->Filename, $this->Path, $this->Extension, $this->getElementI('origwidth'), $this->getElementI('origheight'), '', null, $thumb['Quality']);

			if($thumbObj->exists() && $thumbObj->getOutputPath() != $this->Path){
				$this->Thumbs[] = $thumb['ID'];
			}
		}

		return $this->Thumbs;
	}

	/**
	 * returns the path for the template to be included
	 *
	 * @return string
	 */
	function editor(we_base_jsCmd $jsCmd = null): we_editor_base{
		switch($this->EditPageNr){
			case we_base_constants::WE_EDITPAGE_THUMBNAILS:
				return new we_editor_thumbnails($this, $jsCmd);
			default:
				return parent::editor($jsCmd);
		}
	}

	/**
	 * adds thumbnails to the image document
	 *
	 * @param int[] $thumbsToAdd
	 */
	function add_thumbnails(array $thumbsToAdd): void{
		$this->Thumbs = array_unique(array_filter(array_merge($this->Thumbs, $thumbsToAdd)));
		$this->DocChanged = true;
	}

	/**
	 * deletes a thumbnail from the image document
	 *
	 * @return void
	 * @param int $thumbnailID
	 */
	function del_thumbnails(int $thumbnailID): void{
		$pos = array_search($thumbnailID, $this->Thumbs);
		if($pos !== false){
			unset($this->Thumbs[$pos]);
		}
		$this->DocChanged = true;
	}

	/**
	 * sets extra attributes for the image
	 *
	 * @return void
	 * @param array<string,mixed> $attribs
	 */
	function initByAttribs(array $attribs): void{
		foreach($attribs as $a => $b){
			if(strtolower($a) != 'id' && $b != ''){
				if((array_key_exists($a, self::ELEMENT_KEY) ? self::ELEMENT_KEY[$a] : self::ETYPE_TEXT) === self::ETYPE_NUMBER){
					$this->setElementI($a, (int) $b, 'attrib');
				} elseif(is_string($b)){
					$this->setElementS($a, $b, 'attrib');
				} elseif(is_bool($b)){
					$this->setElementB($a, $b, 'attrib');
				} else {
					t_e('attrib unknown', $a, $b);
				}
			}
		}
		$this->checkDisableEditpages();
	}

	public function initByID(int $id, string $table = '', int $from = self::LOAD_MAIN_DB): void{
		if(!$id){
			$this->we_new();
			return;
		}
		parent::initByID($id, $table, $from);
		if(isWE() && we_base_WE::$editmode){
			$this->checkDisableEditpages();
		}
	}

	public static function isSvg(string $extension): bool{
		return in_array($extension, self::SVG_EXTENSION);
	}

	/**
	 * returns the javascript for the rollover function
	 *
	 * @return string
	 */
	function getRollOverScript(string $parent = 'span'): string{
		if(!$this->getElementB('RollOverFlag')){
			return '';
		}

		if(!$this->getElementS('name')){
			$this->setElementS('name', 'ro_' . $this->Name, 'attrib');
		}
		$class = $this->getRollOverClass();
		return we_html_element::cssElement('
' . $parent . ' img.rolloverOut{display:none;}
' . $parent . '.' . $class . ':hover img.rolloverIn{
	display:none;
}
' . $parent . '.' . $class . ':hover img.rolloverOut{
	display:initial;
}');
	}

	function getRollOverUrl(): string{
		return fS('SELECT Path FROM ' . FILE_TABLE . ' WHERE ID=' . $this->getElementI('RollOverID'));
	}

	/**
	 * @return array
	 * @desc returns the rollover attribs as array
	 */
	function getRollOverClass(): string{
		return $this->getElementB('RollOverFlag') ? 'rollover_' . self::$imgCnt : '';
	}

	/**
	 * resizes the image with the new $width & $height
	 *
	 * @return boolean
	 * @param int $width
	 * @param int $height
	 * @param int $quality
	 * @param bool $ratio
	 */
	function resizeImage(int $width, int $height, int $quality = 8, bool $ratio = false): bool{
		if(self::NEW_SVG_ACTIVE && self::isSvg($this->Extension)){ //@phan-suppress-current-line PhanImpossibleCondition
			$this->setElementI('width', $width);
			$this->setElementI('height', $height);
			$this->setElementS('scale', 'scale(.1)', 'attrib');
		} else {
			$quality = (int) (($quality > 10 ? 10 : max($quality, 0)) * 10);
			$dataPath = TEMP_PATH . we_base_file::getUniqueId();
			we_base_file::insertIntoCleanUp($dataPath, 3600);
			$resized_image = we_base_imageEdit::edit_image($this->getElementS('data'), $this->getGDType(), $dataPath, $quality, $width, $height, [$ratio ? we_contents_thumbnail::OPTION_RATIO : 0]);
			if(!$resized_image[0]){
				return false;
			}
			$this->setElementS('data', $dataPath);

			$this->setElementI('width', $resized_image[1], 'attrib');
			$this->setElementI('origwidth', $resized_image[1], 'attrib');

			$this->setElementI('height', $resized_image[2], 'attrib');
			$this->setElementI('origheight', $resized_image[2], 'attrib');
		}

		$this->DocChanged = true;
		return true;
	}

	/**
	 * rotates the image with the new $width, $height and rotation angle
	 *
	 * @return boolean
	 * @param int $width
	 * @param int $height
	 * @param int $rotation
	 * @param int $quality
	 */
	function rotateImage(int $width, int $height, int $rotation, int $quality = 8): bool{
		if(self::NEW_SVG_ACTIVE && self::isSvg($this->Extension)){ //@phan-suppress-current-line PhanImpossibleCondition
			$matches = [];
			preg_match('|rotate\((\d+)deg\)|', $this->getElementS('rotate'), $matches);
			$deg = (int) isset($matches[1]) ? $matches[1] : 0;
			preg_match('|scale\([-+]?(\d*\.?\d*)\)|', $this->getElementS('scale'), $matches);
			$quot = (float) isset($matches[1]) ? $matches[1] : 1;

			$rotation = $rotation === 90 ? 270 : ($rotation === 270 ? 90 : $rotation);
			$this->setElementS('rotate', (($deg = ($deg + $rotation) % 360) ? 'rotate(' . $deg . 'deg)' : ''), 'attrib');

			if($rotation % 180){
				$quot = round($quot * ($width = $this->getElementI('width')) / (($height = $this->getElementI('height')) ?: 1), (int) ceil(4 - log10(min([$width, $height]))));
				$tmp = $this->getElementI('width');
				$this->setElementI('width', $this->getElementI('height'), 'attrib');
				$this->setElementI('height', $tmp, 'attrib');
				[$v1, $v2, $v3, $v4] = explode(' ', $this->getElementS('viewbox'));
				$this->setElementS('viewbox', implode(' ', [$v1, $v2, $v3, $v4]), 'attrib'); //
			}
			$this->setElementS('scale', ($quot !== 1.0 ? 'scale(' . $quot . ')' : ''), 'attrib');
		} else {
			$quality = (int) (max(min($quality, 10), 0) * 10);

			$dataPath = TEMP_PATH . we_base_file::getUniqueId();
			we_base_file::insertIntoCleanUp($dataPath, 3600);
			$resized_image = we_base_imageEdit::edit_image($this->getElementS('data'), $this->getGDType(), $dataPath, $quality, $width, $height, [we_contents_thumbnail::OPTION_INTERLACE], [
					0, 0], $rotation);

			if(!$resized_image[0]){
				return false;
			}
			$this->setElementS('data', $dataPath);

			$this->setElementI('width', $resized_image[1], 'attrib');
			$this->setElementI('origwidth', $resized_image[1], 'attrib');

			$this->setElementI('height', $resized_image[2], 'attrib');
			$this->setElementI('origheight', $resized_image[2], 'attrib');
		}

		$this->DocChanged = true;
		return true;
	}

	public function fileExists(): bool{
		return file_exists(WEBEDITION_PATH . '../' . $this->Path) && is_readable(WEBEDITION_PATH . '../' . $this->Path) && !is_dir(WEBEDITION_PATH . '../' . $this->Path);
	}

	/**
	 * gets the HTML for including in HTML-Docs.
	 * If a thumbnail should displayed and it doesn't exists,
	 * it will be created automatically
	 *
	 * @return string
	 * @param boolean $dyn
	 * @param bool $inc_href
	 */
	function getHtml(bool $dyn = false, bool $inc_href = true, bool $pathOnly = false, string $only = ''): string{
		$data = $this->getElementS('data');
		//if path only - we need to get a possible thumbnail if selected
		$only = ($pathOnly ? 'path' : ($only ?: $this->getElementS('only')));
		$thumbname = $this->getElementS('thumbnail');
		$thumbid = $this->getElementI('thumbnailid');

		switch($only){
			case'id':
				return (string) $this->ID;
			case 'path':
				if(!$thumbname && !$thumbid){
					return $this->Path;
				}
		}

		if($this->ID || ($data && !is_dir($data) && is_readable($data))){
			$img_path = $this->Path;

			// we need to create a thumbnail - check if image exists
			if($img_path && $this->fileExists()){
				if(($thumbname || $thumbid)){
					$thumbObj = new we_contents_thumbnail();
					$ok = ($thumbid ?
						$thumbObj->initByThumbID($thumbid, $this->ID, $this->Filename, $this->Path, $this->Extension, 0, 0) : $thumbObj->initByThumbName($thumbname, $this->ID, $this->Filename, $this->Path, $this->Extension, 0, 0)
						);
					if($ok){
						$img_path = $thumbObj->getOutputPath();

						if($thumbObj->isOriginal()){
							if(self::isSvg($this->Extension)){
								//FIXME: move this to thumbnail class for all(?) attribs
								$width = $thumbObj->getOutputWidth();
								$height = $thumbObj->getOutputHeight();
								$newstyle = $this->getElementS('style') .
									($width ? ';width:' . $width . 'px' : '') . ($height ? ';height:' . $height . 'px;' : '');
								$this->setElementS('style', $newstyle, 'attrib');
								$this->delElement('width');
								$this->delElement('height');
							}
						} elseif(
						// open a file
							file_exists(WEBEDITION_PATH . '..' . $img_path) &&
							filectime(WEBEDITION_PATH . '..' . $img_path) > $thumbObj->getDate()){//picture created after thumbnail definition was changed, so all is up-to-date
						} else {
							$thumbObj->getThumb();
						}

						if($this->getElementI('width')){
							$this->setElementI('width', $thumbObj->getOutputWidth(), 'attrib');
						}
						if($this->getElementI('height')){
							$this->setElementI('height', $thumbObj->getOutputHeight(), 'attrib');
						}
					}
				} else {
					// IMI SVG: may be wrong!
					if(self::NEW_SVG_ACTIVE && self::isSvg($this->Extension) && !$this->getElementI('width') && !$this->getElementI('height')){//@phan-suppress-current-line PhanImpossibleCondition
						$this->setElementI('width', 100);
					}
				}

				$src = $dyn ?
					WEBEDITION_DIR . 'we_cmd.php?we_cmd[0]=we_base_showBinaryDoc&we_cmd[1]=' . $this->ContentType . '&we_cmd[2]=' . we_base_WE::$transaction . '&rand=' . microtime() :
					$img_path . (DISABLE_MEDIA_MODTIME ? '' : '?m=' . ($this->Publish_Date ? $this->Publish_Date->format('U') : $this->Mod_Date->format('U')));
			} else {
				$img_path = '';
				$src = ICON_DIR . 'no_image.gif';
			}
			switch($only){
				case 'path':
					return $img_path;
				case 'src':
					return $src;
			}

			switch($this->getElementS('LinkType')){
				case we_base_link::TYPE_INT:
					$href = fS('SELECT Path FROM ' . FILE_TABLE . ' WHERE ID=' . $this->getElementI('LinkID'));
					break;
				case we_base_link::TYPE_EXT:
					$href = $this->getElementS('LinkHref');
					break;
				case we_base_link::TYPE_OBJ:
					$id = $this->getElementI('ObjID');
					if(!we_base_WE::$mainDoc['doc']){
						$pidCvs = fS('SELECT Workspaces FROM ' . OBJECT_FILES_TABLE . ' WHERE ID=' . $id);
						$foo = array_filter(explode(',', $pidCvs));
						$pid = (int) ($foo ? $foo[0] : 0);
					} else {
						$pid = we_base_WE::$mainDoc['doc']->ParentID;
					}

					$path = empty(we_base_WE::$mainDoc['doc']) ? '' : we_base_WE::$mainDoc['doc']->Path;
					try{
						$href = we_contents_objectFile::getObjectHref($this->getElementI('ObjID'), $pid, $path, TAGLINKS_DIRECTORYINDEX_HIDE, TAGLINKS_OBJECTSEOURLS, we_base_WE::getWeDoc()->InWebEdition);
//FR #7711
					} catch (we_exception_base $e){//we_link_not_published
						$href = '';
					}
					break;
				default:
					break;
			}

			$target = $this->getElementS('LinkTarget');

			if($this->issetElement('sizingrel')){
				$this->setElementI('width', intval(round($this->getElementI('width') * $this->getElementI('sizingrel'))), 'attrib');
				$this->setElementI('height', intval(round($this->getElementI('height') * $this->getElementI('sizingrel'))), 'attrib');
				$this->delElement('sizingrel');
			}

			if($this->issetElement('sizingbase')){
				$sizingbase = $this->getElementS('sizingbase');
				$this->delElement('sizingbase');
			} else {
				$sizingbase = 16;
			}

			if($this->issetElement('sizingstyle')){
				$tmp = $this->getElementS('sizingstyle');
				$sizingstyle = $tmp === 'none' ? false : $tmp;
				$this->delElement('sizingstyle');
			} else {
				$sizingstyle = false;
			}

			if($sizingstyle){
				$style_width = round($this->getElementI('width') / $sizingbase, 6);
				$style_height = round($this->getElementI('height') / $sizingbase, 6);
				$newstyle = $this->getElementS('style');

				$newstyle .= ';width:' . $style_width . $sizingstyle . ';height:' . $style_height . $sizingstyle . ';';
				$this->setElementS('style', $newstyle, 'attrib');
				$this->delElement('width');
				$this->delElement('height');
			}

			//  Here we generate the image-tag
			//   attribs for the image tag
			$attribs = ['src' => $src];

			$filter = [
				'filesize',
				'type',
				'id',
				'showcontrol',
				'showthumbcontrol',
				'thumbnail',
				'thumbnailid',
				'href',
				'longdescid',
				'showimage',
				'showinputs',
				'listviewname',
				'parentid',
				'startid',
				'origwidth',
				'origheight',
				'useMetaTitle',
				'RollOverFlag',
				'internalwidth',
				'internalheight',
				'internalviewbox',
				'internalpreserveaspectratio',
				'internalrotate',
				'internalscale',
				'viewbox',
				'preserveaspectratio',
				'rotate',
				'scale',
			]; //  dont use these array-entries

			if(defined('HIDENAMEATTRIBINWEIMG_DEFAULT') && HIDENAMEATTRIBINWEIMG_DEFAULT){
				$filter[] = 'name';
			}

			// check longdesc here - does file exist?
			if($this->getElementI('longdescid')){
				$longdesc = id2path($this->getElementI('longdescid'));
				$attribs['longdesc'] = $longdesc;
			}

			// $attribs['Title'] changed to $attribs['title'] to fix bug #5814
			if($this->getElementB('useMetaTitle') && ($title = $this->getElementS('Title'))){ //  set title if set in image
				$attribs['title'] = htmlspecialchars($title);
			}

			foreach($this->nextElement('attrib') as $k){
				'@phan-var string $k';
				if(!in_array($k, $filter, true)){
					$type = (array_key_exists($k, self::ELEMENT_KEY) ? self::ELEMENT_KEY[$k] : self::ETYPE_TEXT);
					$v = ($type === self::ETYPE_NUMBER ? $this->getElementI($k) : $this->getElementS($k));
					if($v){
						$attribs[$k] = $v;
					}
				}
			}

			$showAttrOnly = (empty($attribs['only']) ? '' : $attribs['only']);

			switch($showAttrOnly){
				default:
				case 'src':
				case 'alt':
					return empty($attribs[$showAttrOnly]) ? '' : (string) $attribs[$showAttrOnly];
				case 'filename':
					return $this->Filename;
				case 'size':
					return we_base_file::getHumanFileSize($this->getFilesize());
				case 'extension':
					return $this->Extension;
				case 'parentpath':
					return $this->getParentPath();
				case 'width':
					return (string) ($this->getElementI('width'));
				case 'height':
					return (string) ($this->getElementI('height'));
				case '':
					break;
			}

			if(self::NEW_SVG_ACTIVE && self::isSvg($this->Extension)){//@phan-suppress-current-line PhanImpossibleCondition
				$rotate = ($rot = $this->getElementS('rotate')) && ($rot !== $this->getElementS('internalrotate')) ? $rot : '';
				$scale = ($quot = $this->getElementS('scale')) && ($quot !== $this->getElementS('internalscale')) ? $quot : '';
				$attribs['style'] = ($transform = implode(' ', array_filter([$rotate, $scale]))) ? 'transform:' . $transform . ';' : '' . ($attribs['style'] ?? '');
				$svgview = [];
				$svgview[] = (($box = $this->getElementS('viewbox')) !== $this->getElementS('internalviewbox') ? 'viewBox(' . $box . ')' : '');
				$svgview[] = (($ratio = $this->getElementS('preserveaspectratio')) !== $this->getElementS('internalpreserveaspectratio') ? 'preserveAspectRatio(' . $ratio . ')' : '');
				$attribs['src'] .= (($svgview = array_filter($svgview)) ? '#svgView(' . implode(' ', $svgview) . ')' : '');
			}

			if((!empty($href)) && $inc_href){ //  use link with rollover
				$aAtts = [
					'href' => $href,
					'title' => $attribs['title'] ?? '',
				];

				if($target){
					$aAtts['target'] = $target;
				}
				if(isset($attribs['xml'])){
					$aAtts['xml'] = $attribs['xml'];
				}
				if(($rollOverClass = $this->getRollOverClass())){
					$aAtts['class'] = $rollOverClass;
				}
				$class = (empty($attribs['class']) ? '' : $attribs['class'] . ' ');
				$inImg = we_weTag_tag::htmlTag('img', array_merge($attribs, ['class' => $class . ($rollOverClass ? 'rolloverIn' : '')]));
				$outImg = $rollOverClass ? we_weTag_tag::htmlTag('img', array_merge($attribs, ['class' => $class . 'rolloverOut', 'src' => $this->getRollOverUrl()])) : '';
				return ($this->html = ( $this->getRollOverScript('a') . we_weTag_tag::htmlTag('a', $aAtts, $inImg . $outImg) ));
			}
			if(defined('WE_EDIT_IMAGE')){
				return ($this->html = we_html_element::cssLink(CSS_DIR . 'imageEditTools.css') . we_base_imageCrop::getCrop($this, $attribs));
			}
			$rollOverClass = $this->getRollOverClass();
			$class = (empty($attribs['class']) ? '' : $attribs['class'] . ' ');
			$inImg = we_weTag_tag::htmlTag('img', array_merge($attribs, ['class' => $class . ($rollOverClass ? 'rolloverIn' : '')]));
			$outImg = $rollOverClass ? we_weTag_tag::htmlTag('img', array_merge($attribs, ['class' => $class . 'rolloverOut', 'src' => $this->getRollOverUrl()])) : '';
			return ($this->html = $this->getRollOverScript() .
				($rollOverClass ?
				we_weTag_tag::htmlTag('span', ['class' => $rollOverClass], $inImg . $outImg) :
				$inImg
				)
				);
		}
		if($pathOnly){
			//be compatible
			return '';
		}

		return '';
	}

	/**
	 * function will determine the size of any GIF, JPG, PNG.
	 * This function uses the php Function with the same name.
	 * But the php function doesn't work with some images created from some apps.
	 * So this function uses the gd lib if nothing is returned from the php function
	 *
	 * @static
	 * @return array{int,int,int,string}
	 * @throws we_exception_image
	 * @param $filename complete path of the image
	 */
	function getimagesize(string $filename): array{
		return (self::isSvg($this->Extension) ? self::getSvgSize($filename) : we_contents_thumbnail::getimagesize($filename));
	}

	/**
	 * @param string $filename
	 * @return array{int,int,int,string}
	 */
	private static function getSvgSize(string $filename): array{
		$line = we_base_file::load($filename, 'rb', 1000, true);
		$match = [];
		return [
			(preg_match('|<svg[^>]*width="([^"]*)"[^>]*>|i', $line, $match) ? (int) ($match[1]) : 0),
			(preg_match('|<svg[^>]*height="([^"]*)"[^>]*>|i', $line, $match) ? (int) ($match[1]) : 0),
			0,
			''
		];
	}

	// IMI SVG

	/**
	 * @param string $filename
	 * @return array{int,int,string,string,string,float}
	 */
	private static function extractSvgAttributes(string $filename): array{//FIXME: better return named attribs
		$line = we_base_file::load($filename, 'rb', 1000, true);
		$match = [];
		$viewbox = trim((preg_match('|<svg[^>]*viewBox="([^"]*)"[^>]*>|i', $line, $match) ? (string) ($match[1]) : ''));
		return [
			(int) (preg_match('|<svg[^>]*width="([^"]*)"[^>]*>|i', $line, $match) ? $match[1] : 0),
			(int) (preg_match('|<svg[^>]*height="([^"]*)"[^>]*>|i', $line, $match) ? $match[1] : 0),
			(string) preg_replace('/\s\s+/', ' ', str_replace(',', ' ', trim($viewbox))),
			(string) (preg_match('|<svg[^>]*preserveAspectRatio="([^"]*)"[^>]*>|i', $line, $match) ? $match[1] : ''),
			(string) (preg_match('|<svg[^>]*rotate="([^"]*)"[^>]*>|i', $line, $match) ? $match[1] : ''),
			(float) (preg_match('|<svg[^>]*scale="([^"]*)"[^>]*>|i', $line, $match) ? $match[1] : 0)
			// FIXME: scale and rotate can be found in transform?
		];
	}

	public function isSvgEdited(): bool{
		return $this->getElementI('width') !== $this->getElementI('internalwidth') ||
			$this->getElementI('height') !== $this->getElementI('internalheight') ||
			$this->getElementS('rotate') !== $this->getElementS('internalrotate') ||
			$this->getElementS('viewbox') !== $this->getElementS('internalviewbox') ||
			$this->getElementS('preserveaspectratio') !== $this->getElementS('internalpreserveaspectratio');
	}

	/**
	 * Returns the HTML for the properties part in the properties view
	 *
	 * @return string
	 */
	function formProperties(): string{
		// Create table
		$row = 0;

		if(self::NEW_SVG_ACTIVE){ //@phan-suppress-current-line PhanImpossibleCondition
			$content = (new we_html_table(['class' => 'default propertydualtable'], 5, 3));
			if(self::isSvg($this->Extension)){
				$content->setCol($row, 0, $this->formInput('attrib', 'border#' . self::ETYPE_NUMBER, g_l('weClass', '[border]'), 155, 0, ['type' => 'number', 'onchange' => "_EditorFrame.setEditorIsHot(true);"]))
					->setCol($row++, 1, $this->formInput('attrib', 'align', g_l('weClass', '[align]'), 155, 0, ['onchange' => "_EditorFrame.setEditorIsHot(true);"]))
					->setCol($row, 0, $this->formInput('attrib', 'hspace', g_l('weClass', '[hspace]'), 155, 0, ['onchange' => "_EditorFrame.setEditorIsHot(true);"]))
					->setCol($row++, 1, $this->formInput('attrib', 'vspace', g_l('weClass', '[vspace]'), 155, 0, ['onchange' => "_EditorFrame.setEditorIsHot(true);"]));
			} else {
				$content->setCol($row, 0, $this->formInput('attrib', 'width#' . self::ETYPE_NUMBER, g_l('weClass', '[width]'), 155, 0, ['type' => 'number', 'onchange' => "_EditorFrame.setEditorIsHot(true);"]))
					->setCol($row, 1, $this->formInput('attrib', 'height#' . self::ETYPE_NUMBER, g_l('weClass', '[height]'), 155, 0, ['type' => 'number', 'onchange' => "_EditorFrame.setEditorIsHot(true);"]))
					->setCol($row++, 2, $this->formInput('attrib', 'border#' . self::ETYPE_NUMBER, g_l('weClass', '[border]'), 155, 0, ['type' => 'number', 'onchange' => "_EditorFrame.setEditorIsHot(true);"]))
					->setCol($row, 0, $this->formInput('attrib', 'align', g_l('weClass', '[align]'), 155, 0, ['onchange' => "_EditorFrame.setEditorIsHot(true);"]))
					->setCol($row, 1, $this->formInput('attrib', 'hspace', g_l('weClass', '[hspace]'), 155, 0, ['onchange' => "_EditorFrame.setEditorIsHot(true);"]))
					->setCol($row++, 2, $this->formInput('attrib', 'vspace', g_l('weClass', '[vspace]'), 155, 0, ['onchange' => "_EditorFrame.setEditorIsHot(true);"]));
			}
			$content->setCol($row, 0, $this->formInput('attrib', 'alt', g_l('weClass', '[alt]'), 332, 0, ['onchange' => "_EditorFrame.setEditorIsHot(true);"]), ['colspan' => '2'])
				->setCol($row++, 2, $this->formInput('attrib', 'name', g_l('weClass', '[name]'), 155, 0, ['onchange' => "_EditorFrame.setEditorIsHot(true);"]))
				->setCol($row, 0, $this->formInput('attrib', 'title', g_l('weClass', '[Title]'), 332, 0, ['readonly' => ($this->getElementB('useMetaTitle') ? 'readonly' : ''),
						'onchange' => "_EditorFrame.setEditorIsHot(true);"]), ['colspan' => '2']);
		} else {
			$content = (new we_html_table(['class' => 'default propertydualtable'], 5, 3))
				->setCol($row, 0, $this->formInput('attrib', 'width#' . self::ETYPE_NUMBER, g_l('weClass', '[width]'), 155, 0, ['type' => 'number', 'onchange' => "_EditorFrame.setEditorIsHot(true);"]))
				->setCol($row, 1, $this->formInput('attrib', 'height#' . self::ETYPE_NUMBER, g_l('weClass', '[height]'), 155, 0, ['type' => 'number', 'onchange' => "_EditorFrame.setEditorIsHot(true);"]))
				->setCol($row++, 2, $this->formInput('attrib', 'border#' . self::ETYPE_NUMBER, g_l('weClass', '[border]'), 155, 0, ['type' => 'number', 'onchange' => "_EditorFrame.setEditorIsHot(true);"]))
				// Row 1
				->setCol($row, 0, $this->formInput('attrib', 'align', g_l('weClass', '[align]'), 155, 0, ['onchange' => "_EditorFrame.setEditorIsHot(true);"]))
				->setCol($row, 1, $this->formInput('attrib', 'hspace', g_l('weClass', '[hspace]'), 155, 0, ['onchange' => "_EditorFrame.setEditorIsHot(true);"]))
				->setCol($row++, 2, $this->formInput('attrib', 'vspace', g_l('weClass', '[vspace]'), 155, 0, ['onchange' => "_EditorFrame.setEditorIsHot(true);"]))

				// Row 2
				->setCol($row, 0, $this->formInput('attrib', 'alt', g_l('weClass', '[alt]'), 332, 0, ['onchange' => "_EditorFrame.setEditorIsHot(true);"]), ['colspan' => '2'])
				->setCol($row++, 2, $this->formInput('attrib', 'name', g_l('weClass', '[name]'), 155, 0, ['onchange' => "_EditorFrame.setEditorIsHot(true);"]))
				->setCol($row, 0, $this->formInput('attrib', 'title', g_l('weClass', '[Title]'), 332, 0, ['readonly' => ($this->getElementB('useMetaTitle') ? 'readonly' : ''),
					'onchange' => "_EditorFrame.setEditorIsHot(true);"]), ['colspan' => '2']);
		}

		$titleField = 'we_' . $this->Name . '_attrib[title]';
		//$metaTitleField = 'we_' . $this->Name . '_txt[Title]';
		$useMetaTitle = 'we_' . $this->Name . '_checkAttrib[useMetaTitle#' . self::ETYPE_NUMBER . ']';
		$content->setCol($row++, 2, we_html_forms::checkboxWithHidden($this->getElementB('useMetaTitle'), $useMetaTitle, g_l('weClass', '[use_meta_title]'), false, '', "if(this.checked){ document.forms[0]['" . $titleField . "'].setAttribute('readonly', 'readonly', 'false'); document.forms[0]['" . $titleField . "'].value = '" . $this->getElementS('Title') . "'; }else{ document.forms[0]['" . $titleField . "'].removeAttribute('readonly', 'false'); document.forms[0]['" . $titleField . "'].value='';}_EditorFrame.setEditorIsHot(true);"), [
			'style' => 'vertical-align:bottom']);

		//  longdesc should be available in images.
		//    check if longdesc is set and get path
		$longdesc_id_name = 'we_' . $this->Name . '_attrib[longdescid#' . self::ETYPE_NUMBER . ']';
		$longdesc_text_name = 'tmp_longdesc';
		$longdesc_id = $this->getElementI('longdescid');
		$longdescPath = ($longdesc_id ? id2path($longdesc_id) : '');

		$weSuggest = we_gui_suggest::getInstance();
		$weSuggest->setAcId('LonDesc')
			->setContentType([we_base_ContentTypes::FOLDER_DOCUMENT, we_base_ContentTypes::TEXT, we_base_ContentTypes::WEDOCUMENT, we_base_ContentTypes::HTML])
			->setInput($longdesc_text_name, $longdescPath)
			->setLabel(g_l('weClass', '[longdesc_text]'))
			->setMaxResults(20)
			->setResult($longdesc_id_name, (string) $longdesc_id)
			->setSelector(we_gui_suggest::DocSelector)
			->setWidth(332)
			->setSelectButton(we_html_button::create_button(we_html_button::SELECT, "javascript:we_cmd('we_selector_document',{id:document.we_form.elements['" . $longdesc_id_name . "'].value,table:WE().consts.tables.FILE_TABLE,targetID:'" . $longdesc_id_name . "',targetName:'" . $longdesc_text_name . "',jsCmd:'reload_hot_editpage',filter:WE().consts.contentTypes.WEDOCUMENT+','+WE().consts.contentTypes.TEXT+','+WE().consts.contentTypes.HTML,open_doc:true})"))
			->setTrashButton(we_html_button::create_button(we_html_button::TRASH, "javascript:we_cmd('imageDocument_emptyLongdesk','" . $longdesc_id_name . "','" . $longdesc_text_name . "','" . $weSuggest->getInputId() . "')"));
		$content->setCol($row, 0, $weSuggest->getHTML(), ['style' => 'vertical-align:bottom', 'colspan' => '5']);

		// Return HTML
		return $content->getHtml();
	}

	// IMI SVG
	function formSvgAttribsInternal(): string{
		$row = 0;
		return (new we_html_table(['class' => 'default propertydualtable'], 5, 3))
				->setCol($row, 0, $this->formInput('attrib', 'internalwidth#' . self::ETYPE_NUMBER, g_l('weClass', '[width]'), 155, 0, ['type' => 'number', 'disabled' => 'disabled']))
				->setCol($row, 1, $this->formInput('attrib', 'internalheight#' . self::ETYPE_NUMBER, g_l('weClass', '[height]'), 155, 0, ['type' => 'number', 'disabled' => 'disabled']))
				->setCol($row++, 2, $this->formInput('attrib', 'internalpreserveaspectratio#' . self::ETYPE_TEXT, 'preserveAspectRatio', 155, 0, ['type' => 'text', 'disabled' => 'disabled']))
				->setCol($row, 0, $this->formInput('attrib', 'internalviewbox#' . self::ETYPE_TEXT, 'viewBox', 155, 0, ['type' => 'text', 'disabled' => 'disabled']))
				->setCol($row, 1, $this->formInput('attrib', 'iternalrotate#' . self::ETYPE_TEXT, 'rotate', 155, 0, ['type' => 'text', 'disabled' => 'disabled']))
				->setCol($row, 2, $this->formInput('attrib', 'iternalscale#' . self::ETYPE_TEXT, 'scale', 155, 0, ['type' => 'text', 'disabled' => 'disabled']))
				->getHtml();
	}

	// IMI SVG
	function formSvgAttribsEdited(): string{
		$row = 0;
		return (new we_html_table(['class' => 'default propertydualtable'], 5, 3))
				->setCol($row, 0, $this->formInput('attrib', 'width#' . self::ETYPE_NUMBER, g_l('weClass', '[width]'), 155, 0, ['type' => 'number', 'disabled' => 'disabled']))
				->setCol($row, 1, $this->formInput('attrib', 'height#' . self::ETYPE_NUMBER, g_l('weClass', '[height]'), 155, 0, ['type' => 'number', 'disabled' => 'disabled']))
				->setCol($row++, 2, $this->formInput('attrib', 'preserveaspectratio#' . self::ETYPE_TEXT, 'preserveAspectRatio', 155, 0, ['type' => 'text', 'disabled' => 'disabled']))
				->setCol($row, 0, $this->formInput('attrib', 'viewbox#' . self::ETYPE_TEXT, 'viewBox', 155, 0, ['type' => 'text', 'disabled' => 'disabled']))
				->setCol($row, 1, $this->formInput('attrib', 'rotate#' . self::ETYPE_TEXT, 'rotate', 155, 0, ['type' => 'text', 'disabled' => 'disabled']))
				->setCol($row, 2, $this->formInput('attrib', 'scale#' . self::ETYPE_TEXT, 'scale', 155, 0, ['type' => 'text', 'disabled' => 'disabled']))
				->getHtml();
	}

	/**
	 * Returns true if the gd lib supports the Type of the image
	 * @return boolean
	 */
	function gd_support(): bool{
		return in_array($this->getGDType(), we_base_imageEdit::supported_image_types(), true);
	}

	/**
	 * Returns the Type for the image to use for the gd library functions
	 *
	 * @return string
	 */
	function getGDType(): string{
		return array_key_exists(strtolower($this->Extension), we_base_imageEdit::GDIMAGE_TYPE) ? we_base_imageEdit::GDIMAGE_TYPE[strtolower($this->Extension)] : 'jpg';
	}

	function convert(string $type, int $quality = 8): bool{
		[$width, $height] = $this->getOrigSize();
		$quality = (int) max(min($quality, 10), 0) * 10;

		$dataPath = TEMP_PATH . we_base_file::getUniqueId();
		we_base_file::insertIntoCleanUp($dataPath, 3600);
		we_base_imageEdit::edit_image($this->getElementS('data'), $type, $dataPath, $quality, $width, $height, [we_contents_thumbnail::OPTION_INTERLACE]);

		$this->setElementS('data', $dataPath);
		$this->Extension = '.' . $type;
		$this->Path = $this->getParentPath() . $this->Filename . $this->Extension;

		$this->DocChanged = true;
		return true;
	}

	/**
	 * Creates an html img tag with src to a thumbnail with specified width and height
	 *
	 * @param int $width target width
	 * @param int $height target height
	 * @return string
	 */
	public function getPreviewImg(int $width = 150, int $height = 200): string{
		if(!$this->getElementS('data') || !is_readable($this->getElementS('data'))){
			return $this->getHtml();
		}
		return self::getPreview($this->ID, strtr($this->getElementS('data'), [
				WEBEDITION_PATH => WEBEDITION_DIR,
				WE_DOCUMENT_ROOT => '',
				]), $this->Extension, $width, $height);
	}

	public static function getPreview(int $id, string $path, string $extension, int $width = 150, int $height = 200, array $attribs = [], bool $retSrc = false): string{
		$fullpath = WEBEDITION_PATH . '..' . $path;
		$widthHeight = '';
		$hasWE = str_starts_with($path, WEBEDITION_DIR);
		$exist = file_exists($fullpath);
		if(!$exist){
			$src = ICON_DIR . 'no_image.gif';
			return $retSrc ? $src : we_html_element::img(['src' => $src, 'style' => 'height:64px;width:64px;vertical-align:middle;'] + $attribs);
		}
		$mod = (file_exists($fullpath) ? filemtime($fullpath) : time());
		if(self::isSvg($extension)){
			if($hasWE){
				$path = strtr($path, [WEBEDITION_DIR => '']);
			}
			$src = $hasWE ? WEBEDITION_DIR . 'we_cmd.php?we_cmd[0]=we_base_showTemp&file=' . $path . '&m=' . $mod : $path . '?m=' . $mod;
			$widthHeight = 'width:' . $width . 'px;height:' . $height . 'px;';
		} else {
			$src = WEBEDITION_DIR . 'we_cmd.php?we_cmd[0]=we_base_showThumbnail&' . http_build_query([
					'id' => $id,
					'size' => [
						'width' => $width,
						'height' => $height,
					],
					'path' => $path,
					'extension' => $extension,
					'm' => $mod,
			]);
		}

		return $retSrc ? $src : we_html_element::img([
				'src' => $src,
				'style' => 'max-width:' . $width . 'px;max-height:' . $height . 'px;' . $widthHeight . 'vertical-align:middle;'] + $attribs);
	}

	protected function getMetaDataReader(bool $force = false): we_metadata_metaData{
		return parent::getMetaDataReader(true);
	}

	public function uploadImportFile(array $fileVars, array $docVars): void{
		$tempFile = WE_DOCUMENT_ROOT . $fileVars['fileTemp'];
		if(!self::isSvg($this->Extension) && !in_array(we_base_imageEdit::detect_image_type($tempFile), we_base_imageEdit::GDIMAGE_TYPE, true)){
			throw new we_exception_file(g_l('alert', '[wrong_file][' . $this->ContentType . ']')); //$fileVars['weFileName']
		}

		if($docVars['importMetadata']){
			$this->importMetaData();
		}
		if($docVars['focusX'] || $docVars['focusY']){
			$this->setElementS('focus', '[' . $docVars['focusX'] . ',' . $docVars['focusY'] . ']', 'input');
		}
		try{
			if(self::NEW_SVG_ACTIVE && self::isSvg($this->Extension)){ //@phan-suppress-current-line PhanImpossibleCondition
				$svgAttribs = $this->extractSvgAttributes($tempFile);
				$this->setElementI('width', $svgAttribs[0], 'attrib');
				$this->setElementI('height', $svgAttribs[1], 'attrib');
				$this->setElementS('viewbox', $svgAttribs[2], 'attrib');
				$this->setElementS('preserveaspectratio', $svgAttribs[3], 'attrib');
				$this->setElementS('rotate', $svgAttribs[4], 'attrib');
				$this->setElementS('scale', (string) $svgAttribs[5], 'attrib');

				$this->setElementI('internalwidth', $svgAttribs[0], 'attrib');
				$this->setElementI('internalheight', $svgAttribs[1], 'attrib');
				$this->setElementS('internalviewbox', $svgAttribs[2], 'attrib');
				$this->setElementS('internalpreserveaspectratio', $svgAttribs[3], 'attrib');
				$this->setElementS('internalrotate', $svgAttribs[4], 'attrib');
				$this->setElementS('internalscale', (string) $svgAttribs[5], 'attrib');
			} else {
				$we_size = $this->getimagesize($tempFile);

				if(FILE_UPLOAD_IMG_MAX_SIZE && !we_base_permission::hasPerm([we_base_perm::UPLOAD_DISABLE_DIM_LIMIT]) &&
					max([$we_size[0], $we_size[1]]) > intval(FILE_UPLOAD_IMG_MAX_SIZE)){
					throw new we_exception_base('img longest side error');
				}

				$this->setElementI('width', $we_size[0], 'attrib');
				$this->setElementI('height', $we_size[1], 'attrib');
				$this->setElementI('origwidth', $we_size[0], 'attrib');
				$this->setElementI('origheight', $we_size[1], 'attrib');
			}
		} catch (we_exception_image $e){

		}
		$this->setElementS('title', ($docVars['title'] ?? $this->getElementS('title')), 'attrib');
		$this->setElementS('alt', ($docVars['alt'] ?? $this->getElementS('alt')), 'attrib');
		$this->Thumbs = $docVars['thumbs'] ?? $this->Thumbs;

		parent::uploadImportFile($fileVars, $docVars);
	}

	function importMetaData(array $fieldsToImport = [], bool $importOnlyEmptyFields = false): void{
		$this->getMetaData();
		if(empty($this->metaData) || !is_array($this->metaData)){
			return;
		}

		$fields = [];
		'@phan-var array<string,array> $fields';

		// first we fetch all defined metadata fields from tblMetadata:
		foreach(DB_WE::$inst->getAllq('SELECT tag,type,importFrom FROM ' . METADATA_TABLE) as $record){
			['tag' => $fieldName, 'type' => $fieldType, 'importFrom' => $importFrom] = $record;
			$fieldType = $fieldType ?: 'textfield';

			$parts = explode(',', $importFrom);
			foreach($parts as $part){
				$part = trim($part);
				$fieldParts = explode('/', $part);
				if(count($fieldParts) > 1){
					$tagType = strtolower(trim($fieldParts[0]));
					$tagName = trim($fieldParts[1]);
					if(!(isset($fields[$fieldName]) && is_array($fields[$fieldName]))){
						$fields[$fieldName] = [];
					}
					$fields[$fieldName][] = [$tagType, $tagName, $fieldType];
				}
			}
		}

		$typeMap = ['textfield' => 'txt', 'wysiwyg' => 'txt', 'textarea' => 'txt', 'date' => 'date'];
		$regs = [];

		foreach($fields as $fieldName => $arr){
			$fieldName = (string) $fieldName;
			$fieldVal = $this->getElementS($fieldName);

			if((empty($fieldsToImport) || in_array($fieldName, array_keys($fieldsToImport), true)) && (!$importOnlyEmptyFields || $fieldVal === '')){
				foreach($arr as $impFr){
					if(isset($this->metaData[$impFr[0]][$impFr[1]]) && !empty($this->metaData[$impFr[0]][$impFr[1]])){
						$val = (string) ($this->metaData[$impFr[0]][$impFr[1]]);
						if($impFr[2] === 'date'){
							// here we need to parse the date
							if(preg_match('|^(\d{4}):(\d{2}):(\d{2}) (\d{2}):(\d{2}):(\d{2})$|', $val, $regs)){
								$val = sprintf('%016d', mktime((int) $regs[4], (int) $regs[5], (int) $regs[6], (int) $regs[2], (int) $regs[3], (int) $regs[1]) ?: 0);
							}
						}
						$this->setElementS($fieldName, trim($val), $typeMap[$impFr[2]]);
						break;
					}
				}
			}
		}
	}

	/**
	 * Returns the HTML for the link part in the properties view
	 *
	 * @return string
	 */
	function formLink(): string{
		$weSuggest = we_gui_suggest::getInstance();

		$linkType = $this->getElementS('LinkType') ?: 'no';
		$checkLinkname = 'we_' . $this->Name . '_txt[LinkType]';

		$row = 0;
// External link
		$textname = 'we_' . $this->Name . '_txt[LinkHref]';
		$btnHrefExt = we_base_permission::hasPerm([we_base_perm::WORK_EXTERNAL_FILES_CAN_SELECT]) ? we_html_button::create_button(we_html_button::SELECT, "javascript:we_cmd('browse_server','" . $textname . "','',document.we_form.elements['" . $textname . "'].value,'check_radio_option," . $checkLinkname . ",1,1')") : '';
		$ext_link_table = (new we_html_table(['class' => 'default'], 1, 2))
			->setCol(0, 0, we_html_tools::textInput('we_' . $this->Name . '_txt[LinkHref]', $this->getElementS('LinkHref'), 0, ['onchange' => "_EditorFrame.setEditorIsHot(true);"], "text", '280px'))
			->setCol(0, 1, $btnHrefExt);
		$ext_link = 'href' . we_html_element::br() . $ext_link_table->getHtml();

		$content = (new we_html_table(['class' => 'default'], (we_modules_info::isActive(we_modules_info::OBJECT) ? 11 : 9), 2))
			// No link
			// No link
			->setCol($row, 0, we_html_forms::radiobuttonV('no', $linkType, $checkLinkname, g_l('weClass', '[nolink]'), true, '', '_EditorFrame.setEditorIsHot(true);'), [
				'style' => 'vertical-align:top;padding-bottom:10px;'])
			->setCol($row++, 1, '')
			->setCol($row, 0, we_html_forms::radiobuttonV(we_base_link::TYPE_EXT, $linkType, $checkLinkname, g_l('weClass', '[extern]'), true, '', '_EditorFrame.setEditorIsHot(true)'), [
				'style' => 'vertical-align:top;padding-bottom:10px;'])
			->setCol($row++, 1, $ext_link, ['style' => 'vertical-align:top']);

		// Internal link
		$textname = 'we_' . $this->Name . '_vars[LinkPath]';
		$idname = 'we_' . $this->Name . '_link[LinkID#' . self::ETYPE_NUMBER . ']';
		$linkID = $this->getElementI('LinkID', 0, true);
		$linkPath = fS('SELECT Path FROM ' . FILE_TABLE . ' WHERE ID=' . $linkID);
		$weSuggest->setAcId('internalPath')
			->setContentType(array_merge(we_base_ContentTypes::LINK_ABLE, [we_base_ContentTypes::FOLDER_DOCUMENT]))
			->setInput($textname, $linkPath)
			->setResult($idname, (string) $linkID)
			->setTable(FILE_TABLE)
			->setSelectButton(we_html_button::create_button(we_html_button::SELECT, "javascript:we_cmd('we_selector_document', {id:document.we_form.elements['" . $idname . "'].value,table:WE().consts.tables.FILE_TABLE,targetID:'" . $idname . "',targetName:'" . $textname . "',jsCmd:'check_radio_option," . $checkLinkname . ",2,1',rootDirID:0,open_doc:" . (we_base_permission::hasPerm([
						we_base_perm::WORK_WORKSPACE_CAN_SELECT_OTHER_FILES]) ? 0 : 1) . "});"))
			->setMaxResults(10)
			->setWidth(280)
			->setSelector(we_gui_suggest::DocSelector)
			->setLabel('href');
		$int_link = $weSuggest->getHTML();
		$content->setCol($row, 0, we_html_forms::radiobuttonV(we_base_link::TYPE_INT, $linkType, $checkLinkname, g_l('weClass', '[intern]'), true, '', '_EditorFrame.setEditorIsHot(true);'), [
			'style' => 'vertical-align:top']);
		$content->setCol($row++, 1, $int_link, ['style' => 'vertical-align:top']);

		// Object link
		if(we_modules_info::isActive(we_modules_info::OBJECT)){
			$objidname = 'we_' . $this->Name . '_object[ObjID#' . self::ETYPE_NUMBER . ']';
			$objtextname = 'we_' . $this->Name . '_vars[ObjPath]';
			$objPath = fS('SELECT Path FROM ' . OBJECT_FILES_TABLE . ' WHERE ID=' . $this->getElementI('ObjID'));
			$weSuggest->setAcId('objPathLink')
				->setContentType([we_base_ContentTypes::FOLDER_OBJECTFILE, we_base_ContentTypes::OBJECT_FILE])
				->setInput($objtextname, $objPath)
				->setResult($objidname, (string) $this->getElementI('ObjID'))
				->setTable(OBJECT_FILES_TABLE)
				->setSelectButton(we_html_button::create_button(we_html_button::SELECT, "javascript:we_cmd('we_selector_document',{id:document.we_form.elements['" . $objidname . "'].value,table:WE().consts.tables.OBJECT_FILES_TABLE,targetID:'" . $objidname . "',targetName:'" . $objtextname . "',jsCmd:'check_radio_option," . $checkLinkname . ",3,1',filter:WE().consts.contentTypes.OBJECT_FILE,open_doc:" . (we_base_permission::hasPerm([
							we_base_perm::OBJECTS_CAN_SELECT_OTHER_USERS]) ? 0 : 1) . "});"))
				->setMaxResults(10)
				->setWidth(280)
				->setSelector(we_gui_suggest::DocSelector)
				->setLabel('href');
			$obj_link = $weSuggest->getHTML();
			$content
				->setCol($row, 0, we_html_forms::radiobuttonV(we_base_link::TYPE_OBJ, $linkType, 'we_' . $this->Name . '_txt[LinkType]', g_l('linklistEdit', '[objectFile]'), true, '', '_EditorFrame.setEditorIsHot(true);'), [
					'style' => 'vertical-align:top;padding-top:10px;'])
				->setCol($row++, 1, $obj_link, ['style' => 'vertical-align:top']);
		}

		// Target
		$content->setCol($row++, 0, g_l('weClass', '[target]') . we_html_element::br() . we_html_tools::targetBox('we_' . $this->Name . '_txt[LinkTarget]', 0, '', $this->getElementS('LinkTarget'), '_EditorFrame.setEditorIsHot(true);', 97), [
			'colspan' => '2', 'style' => 'vertical-align:top;padding:20px 0;']);

		// Rollover image
		$RollOverFlagName = 'we_' . $this->Name . '_checkAttrib[RollOverFlag#' . self::ETYPE_NUMBER . ']';
		$RollOverFlag = $this->getElementB('RollOverFlag');
		$RollOverIDName = 'we_' . $this->Name . '_link[RollOverID#' . self::ETYPE_NUMBER . ']';
		$RollOverID = $this->getElementI('RollOverID', 0, true);
		$RollOverPathname = 'we_' . $this->Name . '_vars[RollOverPath]';
		$RollOverPath = fS('SELECT Path FROM ' . FILE_TABLE . ' WHERE ID=' . $RollOverID);

		$weSuggest->setAcId('rollOverPath')
			->setContentType(array_merge(we_base_ContentTypes::LINK_ABLE, [we_base_ContentTypes::FOLDER_DOCUMENT]))
			->setInput($RollOverPathname, $RollOverPath)
			->setResult($RollOverIDName, (string) $RollOverID)
			->setTable(FILE_TABLE)
			->setSelectButton(we_html_button::create_button(we_html_button::SELECT, "javascript:we_cmd('we_selector_image',{id:document.we_form.elements['" . $RollOverIDName . "'].value,table:WE().consts.tables.FILE_TABLE,targetID:'" . $RollOverIDName . "',targetName:'" . $RollOverPathname . "',jsCmd:'toggle_checkbox_with_hidden," . $RollOverFlagName . ",1,1',rootDirID:0,open_doc:" . (we_base_permission::hasPerm([
						we_base_perm::WORK_WORKSPACE_CAN_SELECT_OTHER_FILES]) ? 0 : 1) . "});"))
			->setMaxResults(10)
			->setWidth(280)
			->setSelector(we_gui_suggest::DocSelector)
			->setLabel('href');
		$rollover = $weSuggest->getHTML();

		$content
			->setCol($row, 0, we_html_forms::checkboxWithHidden($RollOverFlag, $RollOverFlagName, 'Roll Over', false, '', "we_cmd('setHot')"), ['style' => 'vertical-align:top'])
			->setCol($row, 1, $rollover, ['style' => 'vertical-align:top']);

		return $content->getHtml();
	}

	function hasMetaField(string $name): bool{
		return in_array($name, array_keys(we_metadata_metaData::getDefinedMetaDataFields()), true);
	}

	static function checkAndPrepare(string $formname, we_contents_root $doc): void{
		// check to see if there is an image to create or to change
		if(empty($_FILES['we_ui_' . $formname]) ||
			!is_array($_FILES['we_ui_' . $formname]) ||
			empty($_FILES['we_ui_' . $formname]['name']) ||
			!is_array($_FILES['we_ui_' . $formname]['name'])
		){
			return;
		}

		foreach($_FILES['we_ui_' . $formname]['name'] as $imgName => $filename){
			'@phan-var string $imgName';
			'@phan-var string $filename';
			$imgDataId = we_base_request::string(we_fileupload::USERINPUT_TYPE_IMG . $imgName);

			if(!$imgDataId || !isset($_SESSION[$imgDataId])){
				continue;
			}

			if(we_base_request::bool(we_fileupload::USERINPUT_TYPE_DEL . $imgName)){
				$_SESSION[$imgDataId]['doDelete'] = true;
				$_SESSION[$imgDataId]['id'] = $_SESSION[$imgDataId]['id'] ?: $doc->getElementI($imgName, 0, true);
				continue;
			}
			$_SESSION[$imgDataId]['doDelete'] = false;
			if(!$filename){
				continue;
			}
			// file is selected, check to see if it is an image
			$ct = we_base_ContentTypes::getFromFilename($filename);
			if($ct !== we_base_ContentTypes::IMAGE){
				continue;
			}
			$imgId = $doc->getElementI($imgName);

			// move document from upload location to tmp dir
			$_SESSION[$imgDataId]['serverPath'] = $serverPath = TEMP_PATH . we_base_file::getUniqueId();
			we_base_file::insertIntoCleanUp($serverPath, 3600);
			move_uploaded_file((string) $_FILES['we_ui_' . $formname]['tmp_name'][$imgName], $serverPath);

			try{
				$we_size = we_contents_thumbnail::getimagesize($serverPath);
			} catch (we_exception_image $e){
				unset($_SESSION[$imgDataId]);
				continue;
			}

			$unique = we_base_file::getUniqueId();
			$tmp_Filename = $imgName . '_' . $unique . '_' .
				preg_replace('/[^A-Za-z0-9._-]/', '', (string) $_FILES['we_ui_' . $formname]['name'][$imgName]);

			if($imgId){
				$_SESSION[$imgDataId]['id'] = $imgId;
			}

			$_SESSION[$imgDataId]['fileName'] = preg_replace('#^(.+)\..+$#', '${1}', $tmp_Filename);
			$_SESSION[$imgDataId]['extension'] = (strpos($tmp_Filename, '.') > 0) ? preg_replace('#^.+(\..+)$#', '${1}', $tmp_Filename) : '';
			$_SESSION[$imgDataId]['text'] = $_SESSION[$imgDataId]['fileName'] . $_SESSION[$imgDataId]['extension'];
			$_SESSION[$imgDataId]['unique'] = $unique;

			if((!empty($_SESSION[$imgDataId]['width']) || !empty($_SESSION[$imgDataId]['height'])) &&
				!we_base_imageEdit::isNoThumbnailNeeded($we_size[0],
																						$we_size[1],
																						(int) ($_SESSION[$imgDataId]['width'] ?? 0),
																						(int) ($_SESSION[$imgDataId]['height'] ?? 0),
																						(bool) $_SESSION[$imgDataId]['maximize'])
			){
				$imageData = we_base_file::load($serverPath);
				$thumb = new we_contents_thumbnail();
				$thumb->init(0, (int) $_SESSION[$imgDataId]['width'], (int) $_SESSION[$imgDataId]['height'], [$_SESSION[$imgDataId]['keepratio'] ? we_contents_thumbnail::OPTION_RATIO : 0,
					$_SESSION[$imgDataId]['maximize'] ? we_contents_thumbnail::OPTION_MAXSIZE : 0], '', '', 0, '', $serverPath, (string) $_SESSION[$imgDataId]['extension'], $we_size[0], $we_size[1], $imageData, null, (int) $_SESSION[$imgDataId]['quality']);
				try{
					we_base_file::save($serverPath, $thumb->getThumb());
					$we_size = we_contents_thumbnail::getimagesize($serverPath);
				} catch (we_exception_base $e){
					//do nothing
				}
			}

			$_SESSION[$imgDataId]['imgwidth'] = $we_size[0];
			$_SESSION[$imgDataId]['imgheight'] = $we_size[1];
			$_SESSION[$imgDataId]['type'] = $_FILES['we_ui_' . $formname]['type'][$imgName];
			$_SESSION[$imgDataId]['size'] = $_FILES['we_ui_' . $formname]['size'][$imgName];
		}
	}

	public function getPropertyPage(we_base_jsCmd $jsCmd): string{
		return (new we_html_multiIconBox((new we_html_multiIconBoxContent())
					->add((new we_html_multiIconBoxContentLine(g_l('weClass', '[path]'), $this->formPath(), we_html_multiIconBox::PROP_PATH)))
					->add((new we_html_multiIconBoxContentLine(g_l('weClass', '[document]'), $this->formIsSearchable() . $this->formIsProtected(), we_html_multiIconBox::PROP_DOC)))
					->add((new we_html_multiIconBoxContentLine(g_l('global', '[navigation]'), $this->formNavigation($jsCmd), we_html_multiIconBox::PROP_NAVI))->fold())
					->add(we_modules_info::isActive(we_modules_info::CATEGORY) ? (new we_html_multiIconBoxContentLine(g_l('global', '[categorys]'), $this->formCategory($jsCmd), we_html_multiIconBox::PROP_CATEGORIES))->fold() : null)
					->add((new we_html_multiIconBoxContentLine(g_l('weClass', '[owners]'), $this->formCreatorOwners($jsCmd), we_html_multiIconBox::PROP_USER))->fold())
					->add((new we_html_multiIconBoxContentLine(g_l('weClass', '[hyperlink]'), $this->formLink(), we_html_multiIconBox::PROP_LINK))->fold())
					, 'PropertyPage'))
				->marginLeft()
				->get();
	}

	private function checkDisableEditpages(): void{
		if(self::isSvg($this->Extension)){
			if(($pos = array_search(we_base_constants::WE_EDITPAGE_IMAGEEDIT, $this->EditPageNrs, true)) !== false){
				unset($this->EditPageNrs[$pos]);
			}
			if(($pos = array_search(we_base_constants::WE_EDITPAGE_THUMBNAILS, $this->EditPageNrs, true)) !== false){
				unset($this->EditPageNrs[$pos]);
			}
		}
	}

}
