<?php

declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: d6cd8d745aa0 $
 * $Author: Marc Kraemer <marc.kraemer@webedition.org> $
 * $Date: 2020-12-08 15:19 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */
we_modules_info::isActive(we_modules_info::WORKFLOW);

abstract class we_workflow_utility{

	private static function getTypeForTable(string $table): array{
		switch($table){
			default:
			case FILE_TABLE:
				return [we_workflow_workflow::DOCTYPE_CATEGORY, we_workflow_workflow::FOLDER];
			case OBJECT_FILES_TABLE:
				return [we_workflow_workflow::OBJECT];
		}
	}

	public static function insertDocInWorkflow(int $docID, string $table, int $workflowID, int $userID, string $desc): bool{
		$desc = nl2br($desc);
		$type = self::getTypeForTable($table);
		//create new workflow document
		$doc = we_workflow_document::createNew($docID, $type, $workflowID, $userID);
		$doc->save();
		if(isset($doc->steps[0])){
			$doc->steps[0]->start($desc);
		}
		//insert into document history
		we_workflow_log::logDocumentEvent($doc->ID, $userID, we_workflow_log::TYPE_DOC_INSERTED, $desc);
		$doc->save();
		return true;
	}

	public static function approve(int $docID, string $table, int $userID, string $desc, bool $force = false): bool{
		/* approve step */
		$desc = nl2br($desc);
		$doc = self::getWorkflowDocument($docID, $table, we_workflow_document::STATUS_ACTIVE);
		if(!$doc){
			return false;
		}
		if($doc->approve($userID, $desc, $force)){
			$doc->save();
			return true;
		}
		return false;
	}

	/*

	 */

	public static function decline(int $docID, string $table, int $userID, string $desc, bool $force = false): bool{
		//decline step
		$doc = self::getWorkflowDocument($docID, $table, we_workflow_document::STATUS_ACTIVE);
		if($doc && $doc->ID){
			if($doc->decline($userID, nl2br($desc), $force)){
				$doc->save();
				return true;
			}
		}
		return false;
	}

	/**
	  This function can be used to force removal
	  of document from workflow.
	 */
	public static function removeDocFromWorkflow(int $docID, string $table, int $userID, string $desc): bool{
		$desc = nl2br($desc);
		$doc = self::getWorkflowDocument($docID, $table, we_workflow_document::STATUS_ACTIVE);
		if(!$doc){
			return false;
		}
		if($doc->finishWorkflow(true, $userID)){
			$doc->save();
			//insert into document history
			we_workflow_log::logDocumentEvent($doc->ID, $userID, we_workflow_log::TYPE_DOC_REMOVED, $desc);
			return true;
		}

		return false;
	}

	/**
	  Function returns workflow document object for defined docID
	  If workflow documnet is not defined for that document false
	  will be returned
	 */
	private static function getWorkflowDocument(int $docID, string $table, string $status = we_workflow_document::STATUS_ACTIVE): ?we_workflow_document{
		return we_workflow_document::find($docID, self::getTypeForTable($table), $status);
	}

	/**
	  Same like getWorkflowDocument but returns
	  workflow document id (not object)
	 */
	private static function getWorkflowDocumentID(int $docID, string $table, string $status = we_workflow_document::STATUS_ACTIVE): int{
		$doc = self::getWorkflowDocument($docID, $table, $status);
		return (!empty($doc->ID) ? $doc->ID : 0);
	}

	/**
	 * Functions tries to find workflow for defined
	 * documents parameters and returns new document object
	 * @param int $doctype
	 * @param int[] $categories
	 * @param int $folder
	 * @param int[] $all
	 * @return we_workflow_document
	 */
	public static function getWorkflowDocumentForDoc(int $doctype, array $categories, int $folder, array &$all): we_workflow_document{
		$workflowID = we_workflow_workflow::getDocumentWorkflow($doctype, $categories, $folder, $all);
		$newDoc = new we_workflow_document();
		if(!$workflowID){
			return $newDoc;
		}
		$newDoc->workflowID = $workflowID;
		$newDoc->steps = we_workflow_documentStep::createAllSteps($workflowID);
		return $newDoc;
	}

	/**
	  Functions tries to find workflow for defined
	  objects parametars and returns new document object
	 * @param int $object
	 * @param int[] $categories
	 * @param int $folderID
	 * @param int[] $all
	 * @return we_workflow_document
	 */
	public static function getWorkflowDocumentForObject(int $object, array $categories, int $folderID, array &$all): we_workflow_document{
		$workflowID = we_workflow_workflow::getObjectWorkflow($object, $categories, $folderID, $all);
		$newDoc = new we_workflow_document();
		$newDoc->workflowID = $workflowID;
		$newDoc->steps = we_workflow_documentStep::createAllSteps($workflowID);
		return $newDoc;
	}

	static function getWorkflowName(int $workflowID, string $table): string{
		$foo = self::getAllWorkflows(we_workflow_workflow::ACTIVE, $table);
		return $foo[$workflowID];
	}

	/**
	 *
	 * @param string $workflowName
	 * @param string $table
	 * @return int
	 */
	static function getWorkflowID(string $workflowName, string $table): int{
		$ret = array_search($workflowName, self::getAllWorkflows(we_workflow_workflow::ACTIVE, $table));
		if($ret){
			return $ret;
		}
		throw new we_exception_base('not found');
	}

	/**
	 * @param int $status
	 * @param string $table
	 * @param int[] $all
	 * @return array<int,string>

	 */
	static function getAllWorkflows(int $status = we_workflow_workflow::ACTIVE, string $table = FILE_TABLE, array $all = []): array{ // returns hash array with ID as key and Name as value
		return we_workflow_workflow::getAllWorkflowsInfo($status, self::getTypeForTable($table), $all);
	}

	static function inWorkflow(int $docID, string $table): bool{
		$doc = self::getWorkflowDocument($docID, $table, we_workflow_document::STATUS_ACTIVE);
		return (!empty($doc->ID));
	}

	static function isWorkflowFinished(int $docID, string $table): bool{
		$doc = self::getWorkflowDocument($docID, $table, we_workflow_document::STATUS_ACTIVE);
		if(!$doc){
			return false;
		}
		try{
			$i = $doc->findLastActiveStep();
			return !(($i <= 0) || ($i < count($doc->steps) - 1) || ($doc->steps[$i]->findNumOfFinishedTasks() < count($doc->steps[$i]->tasks)));
		}catch(we_exception_base $e){
			return false;
		}
	}

	/**
	  Function returns true if user is in workflow for
	  defined documnet id, otherwise false
	 */
	static function isUserInWorkflow(int $docID, string $table, int $userID): bool{
		$doc = self::getWorkflowDocument($docID, $table, we_workflow_document::STATUS_ACTIVE);
		if($doc && $doc->ID){
			try{
				$i = $doc->findLastActiveStep();
				$j = $doc->steps[$i]->findTaskByUser($userID);
			}catch(we_exception_base $e){
				return false;
			}

			return ($doc->steps[$i]->tasks[$j]->Status == we_workflow_documentTask::STATUS_NEW);
		}
		return false;
	}

	/**
	  Function returns true if user can edit
	  defined documnet, otherwise false
	 */
	public static function canUserEditDoc(int $docID, string $table, int $userID): bool{
		if(we_base_permission::hasPerm([we_base_perm::ADMINISTRATOR])){
			return true;
		}
		$doc = self::getWorkflowDocument($docID, $table, we_workflow_document::STATUS_ACTIVE);
		if($doc->ID){
			try{
				$i = $doc->findLastActiveStep();
			}catch(we_exception_base $e){
				return false;
			}
			$wStep = new we_workflow_step($doc->steps[$i]->workflowStepID);
			foreach($wStep->tasks as $v){
				if($v->userID == $userID && $v->Edit){
					return true;
				}
			}
		}
		return false;
	}

	/**
	 *
	 * @param int $userID
	 * @param string $table
	 * @param bool $isAdmin
	 * @param bool $permPublish
	 * @param int[] $ws
	 * @return int[]
	 */
	public static function getWorkflowDocsForUser(int $userID, string $table, bool $isAdmin = false, bool $permPublish = false, array $ws = []): array{
		if($isAdmin){
			return self::getAllWorkflowDocs($table);
		}
		$ids = ($permPublish ? self::getWorkflowDocsFromWorkspace($table, $ws) : []);
		$wids = self::getAllWorkflowDocs($table);

		foreach($wids as $id){
			'@phan-var int $id';
			if(!in_array($id, $ids)){
				if(self::isUserInWorkflow($id, $table, $userID)){
					$ids[] = $id;
				}
			}
		}

		return $ids;
	}

	/**
	 *
	 * @param string $table
	 * @return int[]
	 */
	public static function getAllWorkflowDocs(string $table): array{
		$db = DB_WE::$inst;
		$ret = $db->getAllq('SELECT DISTINCT doc.documentID AS ID
FROM ' .
				WORKFLOW_DOC_TABLE . ' doc
WHERE
	doc.Status="' . we_workflow_document::STATUS_ACTIVE . '" AND
	doc.DocumentTable="' . $db->escape(stripTblPrefix($table)) . '"', true);
		'@phan-var int[] $ret';
		return $ret;
	}

	/* public static function getAllUserActiveDocuments(int $intuserID): array{
	  SELECT doc.DocumentTable,doc.documentID,CONCAT_WS(" ",First,Second) AS user,log.Description,COALESCE(f.Path,`of`.Path) AS Path FROM
	  `tblWorkflowTask` t JOIN
	  tblWorkflowDocStep s ON t.stepID=s.workflowStepID JOIN
	  tblWorkflowDoc doc ON s.workFlowDocID=doc.ID JOIN
	  tblWorkflowLog log ON s.ID=log.docStepID JOIN
	  tblUser u ON log.userID=u.ID LEFT JOIN
	  tblFile f ON (f.ID=doc.documentID AND doc.DocumentTable="tblFile") LEFT JOIN
	  tblObjectFiles of (`of`.ID=doc.documentID AND doc.DocumentTable="tblObjectFiles")
	  WHERE t.userID=150 AND t.Edit=1 AND s.Status=0 AND s.startDate>0
	  return [];
	  } */

	/**
	 *
	 * @param string $table
	 * @param int[] $ws
	 * @return int[]
	 */
	private static function getWorkflowDocsFromWorkspace(string $table, array $ws): array{
		$wids = self::getAllWorkflowDocs($table);
		$ids = [];
		foreach($wids as $id){
			'@phan-var int $id';
			if(!in_array($id, $ids)){
				if(!empty($ws)){
					if(we_users_util::in_workspace($id, $ws, $table)){
						$ids[] = $id;
					}
				} else {
					$ids[] = $id;
				}
			}
		}

		return $ids;
	}

	public static function findLastActiveStep(int $docID, string $table): int{
		$doc = self::getWorkflowDocument($docID, $table, we_workflow_document::STATUS_ACTIVE);
		if($doc && $doc->ID){
			return $doc->findLastActiveStep();
		}
		throw new we_exception_base();
	}

	public static function getNumberOfSteps(int $docID, string $table): int{
		$doc = self::getWorkflowDocument($docID, $table, we_workflow_document::STATUS_ACTIVE);
		if($doc && $doc->ID){
			return count($doc->steps);
		}
		throw new we_exception_base();
	}

	public static function getDocumentStatusInfo(int $docID, string $table): string{
		$doc = self::getWorkflowDocumentID($docID, $table);
		return $doc ? we_workflow_view::getDocumentStatus($doc) : '';
	}

	//Cronjob function
	public static function forceOverdueDocuments(int $userID = 0): string{
		$db = new DB_WE();
		$ret = '';
		$all = $db->getAllq('SELECT doc.ID AS docID,docstep.ID AS docstepID,step.ID AS stepID
FROM ' .
				WORKFLOW_DOC_TABLE . ' doc LEFT JOIN ' .
				WORKFLOW_DOC_STEP_TABLE . ' docstep ON doc.ID=docstep.workflowDocID LEFT JOIN ' .
				WORKFLOW_STEP_TABLE . ' step ON docstep.workflowStepID=step.ID
WHERE
	docstep.startDate IS NOT NULL AND
	(docstep.startDate + INTERVAL ROUND(step.Worktime*3600) SECOND)<NOW() AND
	docstep.finishDate IS NULL AND
	docstep.Status="' . we_workflow_documentStep::STATUS_NEW . '" AND
	doc.Status="' . we_workflow_document::STATUS_ACTIVE . '"');
		foreach($all as $cur){
			$workflowDocument = new we_workflow_document($cur['docID']);
			$userID = $userID ?: $workflowDocument->userID;
			$_SESSION['user']['ID'] = $userID;
			if($workflowDocument->document && !self::isWorkflowFinished($workflowDocument->document->ID, $workflowDocument->document->Table)){
				$workflowStep = new we_workflow_step($cur['stepID']);
				if($workflowStep->timeAction){
					$ret .= '(ID: ' . $workflowDocument->ID . ') ';
					if($workflowDocument->findLastActiveStep() >= count($workflowDocument->steps) - 1){
						if($workflowDocument->workflow->LastStepAutoPublish){
							$workflowDocument->autopublish($userID, g_l('modules_workflow', '[auto_published]'), true);
							$ret .= g_l('modules_workflow', '[auto_published]') . "\n";
						} else {
							$workflowDocument->decline($userID, g_l('modules_workflow', '[auto_declined]'), true);
							$ret .= g_l('modules_workflow', '[auto_declined]') . "\n";
						}
					} else {
						$workflowDocument->approve($userID, g_l('modules_workflow', '[auto_approved]'), true);
						$ret .= g_l('modules_workflow', '[auto_approved]') . "\n";
					}
				}
				$workflowDocument->save();
			}
		}
		return $ret;
	}

	public static function getLogButton(int $docID, string $table): string{
		$type = implode(',', self::getTypeForTable($table));
		return we_html_button::create_button(we_html_button::LOG_SHOW, "javascript:(new (WE().layout.overlay)(window,'top', WE().consts.dirs.WEBEDITION_DIR + 'we_cmd.php?we_cmd[0]=we_showMod&mod=workflow&pnt=log&art=" . $docID . "&type=" . $type . "','workflow_history',WE().consts.size.dialog.medium,WE().consts.size.dialog.small)).open();");
	}

	public static function cleanup(): void{
		we_database_base::$inst->query('DELETE FROM ' . WORKFLOW_STEP_TABLE . ' WHERE workflowID NOT IN (SELECT ID FROM ' . WORKFLOW_TABLE . ')');
		we_database_base::$inst->query('DELETE FROM ' . WORKFLOW_DOC_TABLE . ' WHERE workflowID NOT IN (SELECT ID FROM ' . WORKFLOW_TABLE . ')');
		we_database_base::$inst->query('DELETE FROM ' . WORKFLOW_TASK_TABLE . ' WHERE stepID NOT IN (SELECT ID FROM ' . WORKFLOW_STEP_TABLE . ')');
		we_database_base::$inst->query('DELETE FROM ' . WORKFLOW_DOC_STEP_TABLE . ' WHERE workflowStepID NOT IN (SELECT ID FROM ' . WORKFLOW_STEP_TABLE . ') OR workflowDocID NOT IN (SELECT ID FROM ' . WORKFLOW_DOC_TABLE . ')');

		we_database_base::$inst->query('DELETE FROM ' . WORKFLOW_DOC_TASK_TABLE . ' WHERE workflowTaskID NOT IN (SELECT ID FROM ' . WORKFLOW_TASK_TABLE . ') OR documentStepID NOT IN (SELECT ID FROM ' . WORKFLOW_DOC_STEP_TABLE . ')');
		we_database_base::$inst->query('DELETE FROM ' . WORKFLOW_LOG_TABLE . ' WHERE `Date`+INTERVAL 2 YEAR<CURRENT_DATE()');
	}

}
