#!/usr/bin/php
<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: ba441a140530 $
 * $Author: Marc Kraemer <marc.kraemer@webedition.org> $
 * $Date: 2021-01-25 18:28 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */
/*
 * The script makes a rebuild like in the rebuild dialog of webEdition
 * webEdition must be installed
 */
define('WE_DOCUMENT_ROOT', (string) realpath(__DIR__ . '/../../../'));
require_once(WE_DOCUMENT_ROOT . '/webEdition/custom/conf/cliConfig.inc.php'); // @phan-suppress-current-line PhanMissingRequireFile
// CONFIGURATION BEGINS ---------------------------------------------------------

/**
 * Type of Rebuild. Possible Options are:
 *
 * all 			: rebuild all documents and templates
 * templates 	: rebuild all templates
 * static 		: rebuild static documents.
 * objects 		: rebuild all objects
 * navigation 	: rebuild the navigation
 * index		: rebuild the index table
 * thumbnails	: rebuild all thumbnails
 *
 * @var string
 */
$_REQUEST = ['type' => 'all',
	/**
	 * String with comma separated category ids.
	 * If this is set, only documents with the specified
	 * categories will be rebuilt
	 * This is only working when rebuild type is
	 * set to "static"
	 *
	 * @var string
	 */
	'categories' => "",
	/**
	 * Flag if should be an AND instead an OR operation
	 * between the categories.
	 * This is only working when rebuild type is
	 * set to "static"
	 *
	 * @var boolean
	 */
	'catAnd' => false,
	/**
	 * comma separated string with document type ids
	 * If this is set, only documents with the specified
	 * document types will be rebuilt
	 * This is only working when rebuild type is
	 * set to "static"
	 *
	 * @var string
	 */
	'doctypes' => "",
	/**
	 * comma separated string with directory ids.
	 * If this is set, only documents within the specified
	 * directories will be rebuilt
	 * This is only working when rebuild type is
	 * set to "static"
	 *
	 * @var string
	 */
	'directories' => "",
	/**
	 * comma separated string with thumb names to rebuild.
	 *
	 * This needs to be set when rebuild type is
	 * set to "thumbnails"
	 *
	 * @var string
	 */
	'thumbnails' => "",
	/**
	 * If you want to see the output of the script
	 * set this to true;
	 *
	 * @var boolean
	 */
	'verbose' => true,
];

//  END OF OPTIONS


/* #################################### Don't change anything below ############################ */

// we want to see errors
ini_set('display_errors', '1');
error_reporting(E_ALL);


if(!isset($_SERVER['SERVER_NAME'])){
	$_SERVER['SERVER_NAME'] = $SERVER_NAME;
}

//use we-error handler; ignore if logging is disabled!
require_once(WE_DOCUMENT_ROOT . '/webEdition/we/include/we.inc.php');

defined('WE_ERROR_SHOW') || define('WE_ERROR_SHOW', 1);
defined('WE_ERROR_LOG') || define('WE_ERROR_LOG', 1);
we_base_errorHandler::start(false);


// knock out identifiation and permissions
$_SESSION['user']['isWeSession'] = true;
$_SESSION['perms'] = ['ADMINISTRATOR' => true];
$_SESSION['user']['Username'] = 1;


// Define exit codes for errors
define('NO_ARGS', 10);
define('INVALID_OPTION', 11);

// Reading the incoming arguments - same as $argv
$args = Console_Getopt::readPHPArgv();

abstract class rebuild{

	const HELP = 'Usage: rebuild.php [options]
Options:
  -t TYPE, --type=TYPE       Type of rebuild. Possible TYPE values are:
                                * all         : rebuild all documents and templates
                                * templates   : rebuild all templates
                                * static      : rebuild static documents.
                                * objects     : rebuild all objects
                                * thumbnails  : rebuild all thumbnails

  -v, --verbose              Verbosely list files processed
  --help                     Prints out this help

 Options to use when type is set to "all":

Options to use when type is set to "static":

  --categories=CATEGORIES    CATEGORIES is a string with comma separated
                             category ids.
                             If this is set, only documents with the specified
                             categories will be rebuilt

  --catAnd                   if set, all specified categories are ANDed.
                             Only documents which have all categories set are rebuilt.

  --doctypes=DOCTYPES        DOCTYPES is a string with comma separated document type ids
                             If this is set, only documents with the specified
                             document types will be rebuilt

  --directories=DIRECTORIES  DIRECTORIES is a string with comma separated directory ids.
                             If this is set, only documents within the specified
                             directories will be rebuilt

Options to use when type is set to "thumbnails":

  --thumbnails=THUMBNAILS    THUMBNAILS is a comma separated string with
                             thumb names to rebuild.
';

	public static function getOptions($args){
		// Short options
		$short_opts = 'vt:';

// Long options
		$long_opts = [
			'type=',
			'categories=',
			'catAnd',
			'doctypes=',
			'directories=',
			'thumbnails=',
			'verbose',
			'help'
		];

// Convert the arguments to options - check for the first argument
		return ($_SERVER['argv'] && realpath($_SERVER['argv'][0]) == __FILE__ ?
				Console_Getopt::getOpt($args, $short_opts, $long_opts) :
				Console_Getopt::getOpt2($args, $short_opts, $long_opts));
	}

	public static function getRebuildData(): array{
		switch(($type = we_base_request::string('type'))){
			case 'static':
				$_REQUEST['type'] = 'filter';
			//falls through
			case 'all':
			case 'templates':
				return we_rebuild_base::getDocuments('rebuild_' . $type, we_base_request::intList('categories'), we_base_request::bool('catAnd'), we_base_request::intList('doctypes'), we_base_request::intList('directories'));
			case 'objects':
				return we_rebuild_base::getObjects();
			case 'index':
				return we_rebuild_base::getIndex();
			case 'thumbnails':
				$thumbs = DB_WE::$inst->getAllq('SELECT ID FROM ' . THUMBNAILS_TABLE . ' WHERE Name IN ("' . implode('","', array_map('escape_sql_query', we_base_request::stringList('thumbnails'))) . '")');
				'@phan-var int[] $thumbs';
				return we_rebuild_base::getThumbnails($thumbs);
			default:
				print "ERROR: rebuild type is not set!";
				throw new we_exception_base('rebuild type is not set');
		}
	}

}

// Make sure we got them (for non CLI binaries)
if(PEAR::isError($args)){
	fwrite(STDERR, $args->getMessage() . "\n");
	exit(NO_ARGS);
}

$options = rebuild::getOptions($args);

// Check the options are valid
if(PEAR::isError($options)){
	fwrite(STDERR, $options->getMessage() . "\n" . rebuild::HELP . "\n");
	exit(INVALID_OPTION);
}

if(!empty($args)){
	$_REQUEST['verbose'] = false;
	$_REQUEST['catAnd'] = false;
}

foreach($options[0] as $opt){
	switch($opt[0]){
		case '--type':
		case 't':
			$_REQUEST['type'] = $opt[1];
			break;

		case 'v':
		case '--verbose':
			$_REQUEST['verbose'] = true;
			break;

		case '--catAnd':
			$_REQUEST['catAnd'] = true;
			break;


		case '--help':
			print rebuild::HELP;
			exit(0);

		default:
			$_REQUEST[preg_replace('/^--/', '', $opt[0])] = $opt[1];
	}
}

$data = rebuild::getRebuildData();

// start rebuild

foreach($data as $d){
	we_rebuild_base::rebuild($d, we_base_request::bool('verbose'));
}
